from __future__ import annotations

import math
from numbers import Number

import numpy as np
from tlz import first, frequencies

from ..utils import is_integer, parse_bytes

unknown_chunk_message = (
    "\n\n"
    "A possible solution: "
    "https://docs.dask.org/en/latest/array-chunks.html#unknown-chunks\n"
    "Summary: to compute chunks sizes, use\n\n"
    "   x.compute_chunk_sizes()  # for Dask Array `x`\n"
    "   ddf.to_dask_array(lengths=True)  # for Dask DataFrame `ddf`"
)


def broadcast_chunks(*chunkss):
    """Construct a chunks tuple that broadcasts many chunks tuples

    >>> a = ((5, 5),)
    >>> b = ((5, 5),)
    >>> broadcast_chunks(a, b)
    ((5, 5),)

    >>> a = ((10, 10, 10), (5, 5),)
    >>> b = ((5, 5),)
    >>> broadcast_chunks(a, b)
    ((10, 10, 10), (5, 5))

    >>> a = ((10, 10, 10), (5, 5),)
    >>> b = ((1,), (5, 5),)
    >>> broadcast_chunks(a, b)
    ((10, 10, 10), (5, 5))

    >>> a = ((10, 10, 10), (5, 5),)
    >>> b = ((3, 3,), (5, 5),)
    >>> broadcast_chunks(a, b)
    Traceback (most recent call last):
        ...
    ValueError: Chunks do not align: [(10, 10, 10), (3, 3)]
    """
    if not chunkss:
        return ()
    elif len(chunkss) == 1:
        return chunkss[0]
    n = max(map(len, chunkss))
    chunkss2 = [((1,),) * (n - len(c)) + c for c in chunkss]
    result = []
    for i in range(n):
        step1 = [c[i] for c in chunkss2]
        if all(c == (1,) for c in step1):
            step2 = step1
        else:
            step2 = [c for c in step1 if c != (1,)]
        if len(set(step2)) != 1:
            raise ValueError("Chunks do not align: %s" % str(step2))
        result.append(step2[0])
    return tuple(result)


def blockdims_from_blockshape(shape, chunks):
    """

    >>> blockdims_from_blockshape((10, 10), (4, 3))
    ((4, 4, 2), (3, 3, 3, 1))
    >>> blockdims_from_blockshape((10, 0), (4, 0))
    ((4, 4, 2), (0,))
    """
    if chunks is None:
        raise TypeError("Must supply chunks= keyword argument")
    if shape is None:
        raise TypeError("Must supply shape= keyword argument")
    if np.isnan(sum(shape)) or np.isnan(sum(chunks)):
        raise ValueError(
            "Array chunk sizes are unknown. shape: %s, chunks: %s%s"
            % (shape, chunks, unknown_chunk_message)
        )
    if not all(map(is_integer, chunks)):
        raise ValueError("chunks can only contain integers.")
    if not all(map(is_integer, shape)):
        raise ValueError("shape can only contain integers.")
    shape = tuple(map(int, shape))
    chunks = tuple(map(int, chunks))
    return tuple(
        ((bd,) * (d // bd) + ((d % bd,) if d % bd else ()) if d else (0,))
        for d, bd in zip(shape, chunks)
    )


CHUNKS_NONE_ERROR_MESSAGE = """
You must specify a chunks= keyword argument.
This specifies the chunksize of your array blocks.

See the following documentation page for details:
  https://docs.dask.org/en/latest/array-creation.html#chunks
""".strip()


def normalize_chunks(chunks, shape=None, limit=None, dtype=None, previous_chunks=None):
    """Normalize chunks to tuple of tuples

    This takes in a variety of input types and information and produces a full
    tuple-of-tuples result for chunks, suitable to be passed to Array or
    rechunk or any other operation that creates a Dask array.

    Parameters
    ----------
    chunks: tuple, int, dict, or string
        The chunks to be normalized.  See examples below for more details
    shape: Tuple[int]
        The shape of the array
    limit: int (optional)
        The maximum block size to target in bytes,
        if freedom is given to choose
    dtype: np.dtype
    previous_chunks: Tuple[Tuple[int]] optional
        Chunks from a previous array that we should use for inspiration when
        rechunking auto dimensions.  If not provided but auto-chunking exists
        then auto-dimensions will prefer square-like chunk shapes.

    Examples
    --------
    Specify uniform chunk sizes

    >>> from dask.array.core import normalize_chunks
    >>> normalize_chunks((2, 2), shape=(5, 6))
    ((2, 2, 1), (2, 2, 2))

    Also passes through fully explicit tuple-of-tuples

    >>> normalize_chunks(((2, 2, 1), (2, 2, 2)), shape=(5, 6))
    ((2, 2, 1), (2, 2, 2))

    Cleans up lists to tuples

    >>> normalize_chunks([[2, 2], [3, 3]])
    ((2, 2), (3, 3))

    Expands integer inputs 10 -> (10, 10)

    >>> normalize_chunks(10, shape=(30, 5))
    ((10, 10, 10), (5,))

    Expands dict inputs

    >>> normalize_chunks({0: 2, 1: 3}, shape=(6, 6))
    ((2, 2, 2), (3, 3))

    The values -1 and None get mapped to full size

    >>> normalize_chunks((5, -1), shape=(10, 10))
    ((5, 5), (10,))

    Use the value "auto" to automatically determine chunk sizes along certain
    dimensions.  This uses the ``limit=`` and ``dtype=`` keywords to
    determine how large to make the chunks.  The term "auto" can be used
    anywhere an integer can be used.  See array chunking documentation for more
    information.

    >>> normalize_chunks(("auto",), shape=(20,), limit=5, dtype='uint8')
    ((5, 5, 5, 5),)

    You can also use byte sizes (see :func:`dask.utils.parse_bytes`) in place of
    "auto" to ask for a particular size

    >>> normalize_chunks("1kiB", shape=(2000,), dtype='float32')
    ((256, 256, 256, 256, 256, 256, 256, 208),)

    Respects null dimensions

    >>> normalize_chunks((), shape=(0, 0))
    ((0,), (0,))
    """
    if dtype and not isinstance(dtype, np.dtype):
        dtype = np.dtype(dtype)
    if chunks is None:
        raise ValueError(CHUNKS_NONE_ERROR_MESSAGE)
    if isinstance(chunks, list):
        chunks = tuple(chunks)
    if isinstance(chunks, (Number, str)):
        chunks = (chunks,) * len(shape)
    if isinstance(chunks, dict):
        chunks = tuple(chunks.get(i, None) for i in range(len(shape)))
    if isinstance(chunks, np.ndarray):
        chunks = chunks.tolist()
    if not chunks and shape and all(s == 0 for s in shape):
        chunks = ((0,),) * len(shape)

    if (
        shape
        and len(shape) == 1
        and len(chunks) > 1
        and all(isinstance(c, (Number, str)) for c in chunks)
    ):
        chunks = (chunks,)

    if shape and len(chunks) != len(shape):
        raise ValueError(
            "Chunks and shape must be of the same length/dimension. "
            "Got chunks=%s, shape=%s" % (chunks, shape)
        )
    if -1 in chunks or None in chunks:
        chunks = tuple(s if c == -1 or c is None else c for c, s in zip(chunks, shape))

    # If specifying chunk size in bytes, use that value to set the limit.
    # Verify there is only one consistent value of limit or chunk-bytes used.
    for c in chunks:
        if isinstance(c, str) and c != "auto":
            parsed = parse_bytes(c)
            if limit is None:
                limit = parsed
            elif parsed != limit:
                raise ValueError(
                    "Only one consistent value of limit or chunk is allowed."
                    "Used %s != %s" % (parsed, limit)
                )
    # Substitute byte limits with 'auto' now that limit is set.
    chunks = tuple("auto" if isinstance(c, str) and c != "auto" else c for c in chunks)

    if any(c == "auto" for c in chunks):
        chunks = auto_chunks(chunks, shape, limit, dtype, previous_chunks)

    if shape is not None:
        chunks = tuple(c if c not in {None, -1} else s for c, s in zip(chunks, shape))

    if chunks and shape is not None:
        chunks = sum(
            (
                blockdims_from_blockshape((s,), (c,))
                if not isinstance(c, (tuple, list))
                else (c,)
                for s, c in zip(shape, chunks)
            ),
            (),
        )
    for c in chunks:
        if not c:
            raise ValueError(
                "Empty tuples are not allowed in chunks. Express "
                "zero length dimensions with 0(s) in chunks"
            )

    if shape is not None:
        if len(chunks) != len(shape):
            raise ValueError(
                "Input array has %d dimensions but the supplied "
                "chunks has only %d dimensions" % (len(shape), len(chunks))
            )
        if not all(
            c == s or (math.isnan(c) or math.isnan(s))
            for c, s in zip(map(sum, chunks), shape)
        ):
            raise ValueError(
                "Chunks do not add up to shape. "
                "Got chunks=%s, shape=%s" % (chunks, shape)
            )

    return tuple(
        tuple(int(x) if not math.isnan(x) else np.nan for x in c) for c in chunks
    )


def _compute_multiplier(limit: int, dtype, largest_block: int, result):
    """
    Utility function for auto_chunk, to fin how much larger or smaller the ideal
    chunk size is relative to what we have now.
    """
    return (
        limit
        / dtype.itemsize
        / largest_block
        / math.prod(r for r in result.values() if r)
    )


def auto_chunks(chunks, shape, limit, dtype, previous_chunks=None):
    """Determine automatic chunks

    This takes in a chunks value that contains ``"auto"`` values in certain
    dimensions and replaces those values with concrete dimension sizes that try
    to get chunks to be of a certain size in bytes, provided by the ``limit=``
    keyword.  If multiple dimensions are marked as ``"auto"`` then they will
    all respond to meet the desired byte limit, trying to respect the aspect
    ratio of their dimensions in ``previous_chunks=``, if given.

    Parameters
    ----------
    chunks: Tuple
        A tuple of either dimensions or tuples of explicit chunk dimensions
        Some entries should be "auto"
    shape: Tuple[int]
    limit: int, str
        The maximum allowable size of a chunk in bytes
    previous_chunks: Tuple[Tuple[int]]

    See also
    --------
    normalize_chunks: for full docstring and parameters
    """
    if previous_chunks is not None:
        previous_chunks = tuple(
            c if isinstance(c, tuple) else (c,) for c in previous_chunks
        )
    chunks = list(chunks)

    autos = {i for i, c in enumerate(chunks) if c == "auto"}
    if not autos:
        return tuple(chunks)

    if limit is None:
        limit = "128MiB" # config.get("array.chunk-size")
    if isinstance(limit, str):
        limit = parse_bytes(limit)

    if dtype is None:
        raise TypeError("dtype must be known for auto-chunking")

    if dtype.hasobject:
        raise NotImplementedError(
            "Can not use auto rechunking with object dtype. "
            "We are unable to estimate the size in bytes of object data"
        )

    for x in tuple(chunks) + tuple(shape):
        if (
            isinstance(x, Number)
            and np.isnan(x)
            or isinstance(x, tuple)
            and np.isnan(x).any()
        ):
            raise ValueError(
                "Can not perform automatic rechunking with unknown "
                "(nan) chunk sizes.%s" % unknown_chunk_message
            )

    limit = max(1, limit)

    largest_block = math.prod(
        cs if isinstance(cs, Number) else max(cs) for cs in chunks if cs != "auto"
    )

    if previous_chunks:
        # Base ideal ratio on the median chunk size of the previous chunks
        result = {a: np.median(previous_chunks[a]) for a in autos}

        ideal_shape = []
        for i, s in enumerate(shape):
            chunk_frequencies = frequencies(previous_chunks[i])
            mode, count = max(chunk_frequencies.items(), key=lambda kv: kv[1])
            if mode > 1 and count >= len(previous_chunks[i]) / 2:
                ideal_shape.append(mode)
            else:
                ideal_shape.append(s)

        # How much larger or smaller the ideal chunk size is relative to what we have now
        multiplier = _compute_multiplier(limit, dtype, largest_block, result)

        last_multiplier = 0
        last_autos = set()
        while (
            multiplier != last_multiplier or autos != last_autos
        ):  # while things change
            last_multiplier = multiplier  # record previous values
            last_autos = set(autos)  # record previous values

            # Expand or contract each of the dimensions appropriately
            for a in sorted(autos):
                if ideal_shape[a] == 0:
                    result[a] = 0
                    continue
                proposed = result[a] * multiplier ** (1 / len(autos))
                if proposed > shape[a]:  # we've hit the shape boundary
                    autos.remove(a)
                    largest_block *= shape[a]
                    chunks[a] = shape[a]
                    del result[a]
                else:
                    result[a] = round_to(proposed, ideal_shape[a])

            # recompute how much multiplier we have left, repeat
            multiplier = _compute_multiplier(limit, dtype, largest_block, result)

        for k, v in result.items():
            chunks[k] = v
        return tuple(chunks)

    else:
        # Check if dtype.itemsize is greater than 0
        if dtype.itemsize == 0:
            raise ValueError(
                "auto-chunking with dtype.itemsize == 0 is not supported, please pass in `chunks` explicitly"
            )
        size = (limit / dtype.itemsize / largest_block) ** (1 / len(autos))
        small = [i for i in autos if shape[i] < size]
        if small:
            for i in small:
                chunks[i] = (shape[i],)
            return auto_chunks(chunks, shape, limit, dtype)

        for i in autos:
            chunks[i] = round_to(size, shape[i])

        return tuple(chunks)


def round_to(c, s):
    """Return a chunk dimension that is close to an even multiple or factor

    We want values for c that are nicely aligned with s.

    If c is smaller than s we use the original chunk size and accept an
    uneven chunk at the end.

    If c is larger than s then we want the largest multiple of s that is still
    smaller than c.
    """
    if c <= s:
        return max(1, int(c))
    else:
        return c // s * s


def _check_regular_chunks(chunkset):
    """Check if the chunks are regular

    "Regular" in this context means that along every axis, the chunks all
    have the same size, except the last one, which may be smaller

    Parameters
    ----------
    chunkset: tuple of tuples of ints
        From the ``.chunks`` attribute of an ``Array``

    Returns
    -------
    True if chunkset passes, else False

    Examples
    --------
    >>> import dask.array as da
    >>> arr = da.zeros(10, chunks=(5, ))
    >>> _check_regular_chunks(arr.chunks)
    True

    >>> arr = da.zeros(10, chunks=((3, 3, 3, 1), ))
    >>> _check_regular_chunks(arr.chunks)
    True

    >>> arr = da.zeros(10, chunks=((3, 1, 3, 3), ))
    >>> _check_regular_chunks(arr.chunks)
    False
    """
    for chunks in chunkset:
        if len(chunks) == 1:
            continue
        if len(set(chunks[:-1])) > 1:
            return False
        if chunks[-1] > chunks[0]:
            return False
    return True



import numpy as np
from tlz import concat, merge, unique

from cubed.vendor.dask.array.gufunc import _parse_gufunc_signature


def apply_gufunc(
    func,
    signature,
    *args,
    axes=None,
    axis=None,
    output_dtypes=None,
    output_sizes=None,
    vectorize=None,
    **kwargs,
):
    """
    Apply a generalized ufunc or similar python function to arrays.

    This is a cutdown version of the
    `equivalent function <https://docs.dask.org/en/stable/generated/dask.array.gufunc.apply_gufunc.html>`_
    in Dask. Refer there for usage information.

    Current limitations: ``keepdims``, and ``allow_rechunk`` are not supported;
    and multiple outputs are not supported.

    Cubed assumes that ``func`` will allocate a new output array. However, if it allocates more memory
    than than, then you need to tell Cubed about it by setting the ``extra_projected_mem`` parameter
    to the amount needed in bytes (per task).
    """

    # Currently the following parameters cannot be changed
    # keepdims = False
    allow_rechunk = False

    # based on dask's apply_gufunc

    # Input processing:

    # Signature
    if not isinstance(signature, str):
        raise TypeError("`signature` has to be of type string")
    input_coredimss, output_coredimss = _parse_gufunc_signature(signature)

    # Determine nout: nout = None for functions of one direct return; nout = int for return tuples
    nout = None if not isinstance(output_coredimss, list) else len(output_coredimss)

    if nout is not None:
        raise NotImplementedError(
            "Multiple outputs are not yet supported, see https://github.com/cubed-dev/cubed/issues/69"
        )

    if (
        nout is None
        and isinstance(output_dtypes, (tuple, list))
        and len(output_dtypes) == 1
    ):
        output_dtypes = output_dtypes[0]

    # Vectorize function, if required
    if vectorize:
        otypes = output_dtypes
        func = np.vectorize(func, signature=signature, otypes=otypes)

    # Miscellaneous
    if output_sizes is None:
        output_sizes = {}

    # Axes
    # input_axes, output_axes = _validate_normalize_axes(
    #     axes, axis, keepdims, input_coredimss, output_coredimss
    # )

    # Main code:

    # Cast all input arrays to cubed
    # Use a spec if there is one. Note that all args have to have the same spec, and
    # this will be checked later when constructing the plan (see check_array_specs).
    from cubed.array_api.creation_functions import asarray

    specs = [a.spec for a in args if hasattr(a, "spec")]
    spec = specs[0] if len(specs) > 0 else None
    args = [asarray(a, spec=spec) for a in args]

    if len(input_coredimss) != len(args):
        raise ValueError(
            "According to `signature`, `func` requires %d arguments, but %s given"
            % (len(input_coredimss), len(args))
        )

    # Note (cubed): since we don't support allow_rechunk=True, there is no need to transpose args (and outputs back again)

    # Assess input args for loop dims
    input_shapes = [a.shape for a in args]
    input_chunkss = [a.chunks for a in args]
    num_loopdims = [len(s) - len(cd) for s, cd in zip(input_shapes, input_coredimss)]
    max_loopdims = max(num_loopdims) if num_loopdims else None
    core_input_shapes = [
        dict(zip(icd, s[n:]))
        for s, n, icd in zip(input_shapes, num_loopdims, input_coredimss)
    ]
    core_shapes = merge(*core_input_shapes)
    core_shapes.update(output_sizes)

    loop_input_dimss = [
        tuple("__loopdim%d__" % d for d in range(max_loopdims - n, max_loopdims))
        for n in num_loopdims
    ]
    input_dimss = [lp + c for lp, c in zip(loop_input_dimss, input_coredimss)]

    loop_output_dims = max(loop_input_dimss, key=len) if loop_input_dimss else tuple()

    # Assess input args for same size and chunk sizes
    # Collect sizes and chunksizes of all dims in all arrays
    dimsizess = {}
    chunksizess = {}
    for dims, shape, chunksizes in zip(input_dimss, input_shapes, input_chunkss):
        for dim, size, chunksize in zip(dims, shape, chunksizes):
            dimsizes = dimsizess.get(dim, [])
            dimsizes.append(size)
            dimsizess[dim] = dimsizes
            chunksizes_ = chunksizess.get(dim, [])
            chunksizes_.append(chunksize)
            chunksizess[dim] = chunksizes_
    # Assert correct partitioning, for case:
    for dim, sizes in dimsizess.items():
        # Check that the arrays have same length for same dimensions or dimension `1`
        if set(sizes) | {1} != {1, max(sizes)}:
            raise ValueError(f"Dimension `'{dim}'` with different lengths in arrays")
        if not allow_rechunk:
            chunksizes = chunksizess[dim]
            # Check if core dimensions consist of only one chunk
            if (dim in core_shapes) and (chunksizes[0][0] < core_shapes[dim]):
                raise ValueError(
                    "Core dimension `'{}'` consists of multiple chunks. To fix, rechunk into a single \
chunk along this dimension or set `allow_rechunk=True`, but beware that this may increase memory usage \
significantly.".format(dim)
                )
            # Check if loop dimensions consist of same chunksizes, when they have sizes > 1
            relevant_chunksizes = list(
                unique(c for s, c in zip(sizes, chunksizes) if s > 1)
            )
            if len(relevant_chunksizes) > 1:
                raise ValueError(
                    f"Dimension `'{dim}'` with different chunksize present"
                )

    # Apply function - use blockwise here
    arginds = list(concat(zip(args, input_dimss)))

    from cubed.core.ops import blockwise

    # Note (cubed): use blockwise on all output dimensions, not just loop_output_dims like in original
    out_ind = loop_output_dims + output_coredimss

    return blockwise(
        func, out_ind, *arginds, dtype=output_dtypes, new_axes=output_sizes, **kwargs
    )



from dataclasses import dataclass
from typing import List, Optional

from cubed.spec import Spec
from cubed.utils import is_cloud_storage_path


@dataclass
class BufferCopies:
    """Information about the number of buffer copies incurred for array storage operations."""

    read: int
    """The number of copies made when reading an array from storage."""

    write: int
    """The number of copies made when writing an array to storage."""


def get_buffer_copies(spec: Optional[Spec]):
    """Return the number of buffer copies to use, based on the spec.

    Using cloud storage will result in more buffer copies being accounted for.
    """

    # See https://github.com/tomwhite/memray-array
    # More factors (e.g. compression) could be taken into account in the future.

    if (
        spec is not None
        and spec.work_dir is not None
        and is_cloud_storage_path(spec.work_dir)
    ):
        return BufferCopies(read=2, write=2)

    return BufferCopies(read=1, write=1)


def calculate_projected_mem(
    reserved_mem: int,
    inputs: List[int],
    operation: int,
    output: int,
    buffer_copies: BufferCopies,
) -> int:
    """Calculate the projected memory needed to run an array operation.

    All memory sizes are in bytes.

    Parameters
    ----------
    reserved_mem : int
        The memory reserved on a worker for non-data use when running a task, in bytes.
    inputs : list of int
        The sizes of the input arrays, in bytes.
    operation : int
        Extra memory needed for the operation (not including allocating the output), in bytes.
    output : int
        The size of the output array, in bytes.
    buffer_copies: BufferCopies
        The number of buffer copies for reading an writing.

    Returns
    -------
    The projected memory needed to run an array operation.
    """

    projected_mem = reserved_mem

    for input in inputs:
        projected_mem += input * buffer_copies.read
        projected_mem += input

    projected_mem += operation

    projected_mem += output
    projected_mem += output * buffer_copies.write

    return projected_mem


class MemoryModeller:
    """Models peak memory usage for a series of operations."""

    current_mem: int = 0
    peak_mem: int = 0

    def allocate(self, num_bytes):
        self.current_mem += num_bytes
        self.peak_mem = max(self.peak_mem, self.current_mem)

    def free(self, num_bytes):
        self.current_mem -= num_bytes
        self.peak_mem = max(self.peak_mem, self.current_mem)



from cubed.array_api.manipulation_functions import concat


def pad(x, pad_width, mode=None, chunks=None):
    """Pad an array."""
    if len(pad_width) != x.ndim:
        raise ValueError("`pad_width` must have as many entries as array dimensions")
    axis = tuple(
        i
        for (i, (before, after)) in enumerate(pad_width)
        if (before != 0 or after != 0)
    )
    if len(axis) != 1:
        raise ValueError("only one axis can be padded")
    axis = axis[0]
    if pad_width[axis] != (1, 0):
        raise ValueError("only a pad width of (1, 0) is allowed")
    if mode != "symmetric":
        raise ValueError(f"Mode is not supported: {mode}")

    select = []
    for i in range(x.ndim):
        if i == axis:
            select.append(slice(0, 1))
        else:
            select.append(slice(None))
    select = tuple(select)
    a = x[select]
    return concat([a, x], axis=axis, chunks=chunks or x.chunksize)



import numpy as np
import pytest

import cubed
import cubed.array_api as xp
from cubed import apply_gufunc
from cubed._testing import assert_allclose, assert_array_equal
from cubed.backend_array_api import namespace as nxp


@pytest.fixture
def spec(tmp_path):
    return cubed.Spec(tmp_path, allowed_mem=1000000)


@pytest.mark.parametrize("vectorize", [False, True])
def test_apply_reduction(spec, vectorize):
    def stats(x):
        # note dtype matches output_dtypes in apply_gufunc below
        return nxp.mean(x, axis=-1, dtype=np.float32)

    r = np.random.normal(size=(10, 20, 30))
    a = cubed.from_array(r, chunks=(5, 5, 30), spec=spec)
    actual = apply_gufunc(stats, "(i)->()", a, output_dtypes="f", vectorize=vectorize)
    expected = nxp.mean(r, axis=-1, dtype=np.float32)

    assert actual.compute().shape == expected.shape
    assert_allclose(actual.compute(), expected)


def test_apply_gufunc_elemwise_01(spec):
    def add(x, y):
        return x + y

    a = cubed.from_array(np.array([1, 2, 3]), chunks=2, spec=spec)
    b = cubed.from_array(np.array([1, 2, 3]), chunks=2, spec=spec)
    z = apply_gufunc(add, "(),()->()", a, b, output_dtypes=a.dtype)
    assert_array_equal(z, np.array([2, 4, 6]))


def test_apply_gufunc_elemwise_01_non_cubed_input(spec):
    def add(x, y):
        return x + y

    a = cubed.from_array(np.array([1, 2, 3]), chunks=3, spec=spec)
    b = np.array([1, 2, 3])
    z = apply_gufunc(add, "(),()->()", a, b, output_dtypes=a.dtype)
    assert_array_equal(z, np.array([2, 4, 6]))


def test_apply_gufunc_elemwise_loop(spec):
    def foo(x):
        assert x.shape in ((2,), (1,))
        return 2 * x

    a = cubed.from_array(np.array([1, 2, 3]), chunks=2, spec=spec)
    z = apply_gufunc(foo, "()->()", a, output_dtypes=int)
    assert z.chunks == ((2, 1),)
    assert_array_equal(z, np.array([2, 4, 6]))


def test_apply_gufunc_elemwise_core(spec):
    def foo(x):
        assert x.shape == (3,)
        return 2 * x

    a = cubed.from_array(np.array([1, 2, 3]), chunks=3, spec=spec)
    z = apply_gufunc(foo, "(i)->(i)", a, output_dtypes=int)
    assert z.chunks == ((3,),)
    assert_array_equal(z, np.array([2, 4, 6]))


def test_gufunc_two_inputs(spec):
    def foo(x, y):
        return np.einsum("...ij,...jk->ik", x, y)

    a = xp.ones((2, 3), chunks=100, dtype=int, spec=spec)
    b = xp.ones((3, 4), chunks=100, dtype=int, spec=spec)
    x = apply_gufunc(foo, "(i,j),(j,k)->(i,k)", a, b, output_dtypes=int)
    assert_array_equal(x, 3 * np.ones((2, 4), dtype=int))


def test_apply_gufunc_axes_two_kept_coredims(spec):
    ra = np.random.normal(size=(20, 30))
    rb = np.random.normal(size=(10, 1, 40))

    a = cubed.from_array(ra, chunks=(10, 30), spec=spec)
    b = cubed.from_array(rb, chunks=(5, 1, 40), spec=spec)

    def outer_product(x, y):
        return np.einsum("i,j->ij", x, y)

    c = apply_gufunc(outer_product, "(i),(j)->(i,j)", a, b, vectorize=True)
    assert c.compute().shape == (10, 20, 30, 40)


def test_gufunc_output_sizes(spec):
    def foo(x):
        return nxp.broadcast_to(x[:, np.newaxis], (x.shape[0], 3))

    a = cubed.from_array(np.array([1, 2, 3, 4, 5], dtype=int), spec=spec)
    x = apply_gufunc(foo, "()->(i_0)", a, output_dtypes=int, output_sizes={"i_0": 3})
    assert x.chunks == ((5,), (3,))
    assert_array_equal(
        x,
        np.array(
            [
                [1, 1, 1],
                [2, 2, 2],
                [3, 3, 3],
                [4, 4, 4],
                [5, 5, 5],
            ]
        ),
    )



from operator import mul
from typing import Optional, TypeVar

from toolz import map, reduce

from cubed import config
from cubed.backend_array_api import namespace as nxp
from cubed.backend_array_api import numpy_array_to_backend_array
from cubed.runtime.types import Callback, Executor
from cubed.spec import Spec, spec_from_config
from cubed.storage.zarr import open_if_lazy_zarr_array
from cubed.utils import array_memory, itemsize, normalize_chunks

from .plan import arrays_to_plan

sym_counter = 0


def gensym(name="array"):
    """Generate a name with an incrementing counter"""
    global sym_counter
    sym_counter += 1
    return f"{name}-{sym_counter:03}"


# type representing either a CoreArray or a public-facing Array
T_ChunkedArray = TypeVar("T_ChunkedArray", bound="CoreArray")


class CoreArray:
    """Chunked array backed by Zarr storage.

    This class has the basic array attributes (``dtype``, ``shape``, ``chunks``, etc).
    The other array methods and attributes are provided in a subclass.
    """

    def __init__(self, name, zarray, spec, plan):
        self.name = name
        self._zarray = zarray
        self._shape = zarray.shape
        self._dtype = zarray.dtype
        self._chunks = normalize_chunks(
            zarray.chunks, shape=self.shape, dtype=self.dtype
        )
        # get spec from config if not supplied
        self.spec = spec or spec_from_config(config)
        self.plan = plan

    @property
    def blocks(self):
        """An array-like interface to the blocks of an array."""
        from cubed.core.indexing import BlockView

        return BlockView(self)

    @property
    def chunkmem(self):
        """Amount of memory in bytes that a single chunk uses."""
        return array_memory(self.dtype, self.chunksize)

    @property
    def chunksize(self):
        """A tuple indicating the chunk size of each corresponding array dimension."""
        return tuple(max(c) for c in self.chunks)

    @property
    def chunks(self):
        """A tuple containing a sequence of block sizes for each corresponding array dimension."""
        return self._chunks

    @property
    def dtype(self):
        """Data type of the array elements."""
        return self._dtype

    @property
    def ndim(self):
        """Number of array dimensions (axes)."""
        return len(self.shape)

    @property
    def numblocks(self):
        """A tuple indicating the number of blocks (chunks) in each corresponding array dimension."""
        return tuple(map(len, self.chunks))

    @property
    def npartitions(self):
        """Number of chunks in the array."""
        return reduce(mul, self.numblocks, 1)

    @property
    def shape(self):
        """Array dimensions."""
        return self._shape

    @property
    def size(self):
        """Number of elements in the array."""
        return reduce(mul, self.shape, 1)

    @property
    def nbytes(self) -> int:
        """Number of bytes in array"""
        return self.size * itemsize(self.dtype)

    @property
    def itemsize(self) -> int:
        """Length of one array element in bytes"""
        return itemsize(self.dtype)

    def _read_stored(self):
        # Only works if the array has been computed
        if self.size > 0:
            # read back from zarr
            zarray = open_if_lazy_zarr_array(self._zarray)
            return numpy_array_to_backend_array(zarray[...])
        else:
            # this case fails for zarr, so just return an empty array of the correct shape
            return nxp.empty(self.shape, dtype=self.dtype)

    def compute(
        self,
        *,
        executor=None,
        callbacks=None,
        optimize_graph=True,
        optimize_function=None,
        resume=None,
        **kwargs,
    ):
        """Compute this array, and any arrays that it depends on.

        Parameters
        ----------
        executor : Executor, optional
            The executor to use to run the computation. If not set then the executor
            from the array's ``Spec`` is used.
        callbacks : list of Callback, optional
            Callbacks to send events to while running the computation.
        optimize_graph : bool, optional
            If True, the graph is optimized before running the computation.
            Otherwise, the graph is computed as is. Default is True.
        optimize_function : callable, optional
            Function to optimize the computation graph. Default is to use a
            function that performs map fusion only.
        resume : bool, optional
            If True, intermediate arrays that have already been computed won't be
            recomputed. Default is False.
        """
        result = compute(
            self,
            executor=executor,
            callbacks=callbacks,
            optimize_graph=optimize_graph,
            optimize_function=optimize_function,
            resume=resume,
            **kwargs,
        )
        if result:
            return result[0]

    def rechunk(
        self: T_ChunkedArray, chunks, *, min_mem=None, use_new_impl=True
    ) -> T_ChunkedArray:
        """Change the chunking of this array without changing its shape or data.

        Parameters
        ----------
        chunks : tuple
            The desired chunks of the array after rechunking.

        Returns
        -------
        cubed.CoreArray
            An array with the desired chunks.
        """
        from cubed.core.ops import rechunk

        return rechunk(self, chunks, min_mem=min_mem, use_new_impl=use_new_impl)

    def visualize(
        self,
        filename="cubed",
        format=None,
        optimize_graph=True,
        optimize_function=None,
        show_hidden=False,
    ):
        """Produce a visualization of the computation graph for this array.

        Parameters
        ----------
        filename : str
            The name of the file to write to disk. If the provided ``filename``
            doesn't include an extension, '.svg' will be used by default.
        format : {'png', 'pdf', 'dot', 'svg', 'jpeg', 'jpg'}, optional
            Format in which to write output file.  Default is 'svg'.
        optimize_graph : bool, optional
            If True, the graph is optimized before rendering.  Otherwise,
            the graph is displayed as is. Default is True.
        optimize_function : callable, optional
            Function to optimize the computation graph. Default is to use a
            function that performs map fusion only.
        show_hidden : bool
            If True, show arrays and operations that are marked as hidden in the
            visualization. Default is False.

        Returns
        -------
        IPython.display.SVG, or None
            An IPython SVG image if IPython can be imported (for rendering
            in a notebook), otherwise None.
        """
        return visualize(
            self,
            filename=filename,
            format=format,
            optimize_graph=optimize_graph,
            optimize_function=optimize_function,
            show_hidden=show_hidden,
        )

    def __getitem__(self: T_ChunkedArray, key, /) -> T_ChunkedArray:
        from cubed.core.indexing import index

        return index(self, key)

    def __repr__(self):
        return f"cubed.core.CoreArray<{self.name}, shape={self.shape}, dtype={self.dtype}, chunks={self.chunks}>"


def check_array_specs(arrays):
    specs = [a.spec for a in arrays if hasattr(a, "spec")]
    if not all(s == specs[0] for s in specs):
        raise ValueError(
            f"Arrays must have same spec in single computation. Specs: {specs}"
        )
    return arrays[0].spec


def compute(
    *arrays,
    executor=None,
    callbacks=None,
    optimize_graph=True,
    optimize_function=None,
    resume=None,
    **kwargs,
):
    """Compute multiple arrays at once.

    Parameters
    ----------
    arrays : cubed.CoreArray
        The arrays to compute.
    executor : Executor, optional
        The executor to use to run the computation. If not set then the executor
        from the ``Spec`` from the arrays is used.
    callbacks : list of Callback, optional
        Callbacks to send events to while running the computation.
    optimize_graph : bool, optional
        If True, the graph is optimized before running the computation.
        Otherwise, the graph is computed as is. Default is True.
    optimize_function : callable, optional
        Function to optimize the computation graph. Default is to use a
        function that performs map fusion only.
    resume : bool, optional
        If True, intermediate arrays that have already been computed won't be
        recomputed. Default is False.
    """
    spec = check_array_specs(arrays)  # guarantees all arrays have same spec
    plan = arrays_to_plan(*arrays)
    if executor is None:
        executor = arrays[0].spec.executor
        if executor is None:
            from cubed.runtime.executors.local import ThreadsExecutor

            executor = ThreadsExecutor()

    # combine any callbacks specified as args with any active callbacks from the context manager
    if callbacks is None and len(Callback.active) == 0:
        all_callbacks = None
    else:
        all_callbacks = list(Callback.active)
        if callbacks is not None:
            all_callbacks.extend(callbacks)

    _return_in_memory_array = kwargs.pop("_return_in_memory_array", True)
    plan.execute(
        executor=executor,
        callbacks=all_callbacks,
        optimize_graph=optimize_graph,
        optimize_function=optimize_function,
        resume=resume,
        array_names=tuple(a.name for a in arrays),
        spec=spec,
        **kwargs,
    )

    if _return_in_memory_array:
        return tuple(a._read_stored() for a in arrays)


def visualize(
    *arrays,
    filename="cubed",
    format=None,
    optimize_graph=True,
    optimize_function=None,
    show_hidden=False,
):
    """Produce a visualization of the computation graph for multiple arrays.

    Parameters
    ----------
    arrays : cubed.CoreArray
        The arrays to include in the visualization.
    filename : str
        The name of the file to write to disk. If the provided ``filename``
        doesn't include an extension, '.svg' will be used by default.
    format : {'png', 'pdf', 'dot', 'svg', 'jpeg', 'jpg'}, optional
        Format in which to write output file.  Default is 'svg'.
    optimize_graph : bool, optional
        If True, the graph is optimized before rendering.  Otherwise,
        the graph is displayed as is. Default is True.
    optimize_function : callable, optional
        Function to optimize the computation graph. Default is to use a
        function that performs map fusion only.
    show_hidden : bool
        If True, show arrays and operations that are marked as hidden in the
        visualization. Default is False.

    Returns
    -------
    IPython.display.SVG, or None
        An IPython SVG image if IPython can be imported (for rendering
        in a notebook), otherwise None.
    """
    plan = arrays_to_plan(*arrays)
    return plan.visualize(
        filename=filename,
        format=format,
        optimize_graph=optimize_graph,
        optimize_function=optimize_function,
        show_hidden=show_hidden,
        array_names=tuple(a.name for a in arrays),
    )


class PeakMeasuredMemoryCallback(Callback):
    def on_task_end(self, event):
        self.peak_measured_mem = event.peak_measured_mem_end


def measure_reserved_mem(
    executor: Executor, work_dir: Optional[str] = None, **kwargs
) -> int:
    """Measures the reserved memory use for a given executor runtime.

    This is the memory used by the Python process for running a task,
    excluding any memory used for data for the computation. It can vary by
    operating system, Python version, executor runtime, and installed package
    versions.

    It can be used as a guide to set ``reserved_mem`` when creating a ``Spec`` object.

    This implementation works by running a trivial computation on a tiny amount of
    data and measuring the peak memory use.

    Parameters
    ----------
    executor : cubed.runtime.types.Executor
        The executor to use to run the computation. It must be an executor that
        reports peak memory, such as Lithops or Modal.

    work_dir : str or None, optional
        The directory path (specified as an fsspec or obstore URL) used for storing intermediate data.
        This is required when using a cloud runtime.

    kwargs
        Keyword arguments to pass to the ``compute`` function.

    Returns
    -------
    int
        The memory used by the runtime in bytes.
    """
    import cubed.array_api as xp

    # give a generous memory allowance
    a = xp.ones((1,), spec=Spec(work_dir, allowed_mem="500MB"))
    b = xp.negative(a)
    peak_measured_mem_callback = PeakMeasuredMemoryCallback()
    b.compute(executor=executor, callbacks=[peak_measured_mem_callback], **kwargs)
    return peak_measured_mem_callback.peak_measured_mem



import numpy as np
import pytest

import cubed
import cubed.array_api as xp
from cubed._testing import assert_allclose, assert_array_equal
from cubed.array_api.manipulation_functions import reshape_chunks
from cubed.backend_array_api import namespace as nxp
from cubed.tests.utils import (
    ALL_EXECUTORS,
    MAIN_EXECUTORS,
    MODAL_EXECUTORS,
    skip_if_cupy,
)


@pytest.fixture
def spec(tmp_path):
    return cubed.Spec(tmp_path, allowed_mem=100000)


@pytest.fixture(
    scope="module",
    params=MAIN_EXECUTORS,
    ids=[executor.name for executor in MAIN_EXECUTORS],
)
def executor(request):
    if request.param.name == "processes" and "cupy" in nxp.__name__:
        pytest.skip(reason="CuPy is not supported with 'processes' executor")
    return request.param


@pytest.fixture(
    scope="module",
    params=ALL_EXECUTORS,
    ids=[executor.name for executor in ALL_EXECUTORS],
)
def any_executor(request):
    if request.param.name == "processes" and "cupy" in nxp.__name__:
        pytest.skip(reason="CuPy is not supported with 'processes' executor")
    return request.param


@pytest.fixture(
    scope="module",
    params=MODAL_EXECUTORS,
    ids=[executor.name for executor in MODAL_EXECUTORS],
)
def modal_executor(request):
    return request.param


# Array object


def test_object_bool(tmp_path, executor):
    spec = cubed.Spec(tmp_path, 100000, executor=executor)
    a = xp.asarray(
        [[False, False, False], [False, False, False], [False, False, False]],
        chunks=(2, 2),
        spec=spec,
    )
    b = xp.all(a)
    assert not b

    a = xp.asarray(
        [[True, True, True], [True, True, True], [True, True, True]],
        chunks=(2, 2),
        spec=spec,
    )
    b = xp.all(a)
    assert b


def test_object_transpose(spec):
    a = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 2), spec=spec)

    assert_array_equal(a.mT.compute(), np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]).T)

    assert_array_equal(a.T.compute(), np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]).T)


def test_reflected_operation(spec):
    a = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 2), spec=spec)
    b = 1 - a

    assert_array_equal(b.compute(), 1 - np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]))


# Creation functions


def test_arange(spec, executor):
    a = xp.arange(12, chunks=(5,), spec=spec)
    assert_array_equal(a.compute(executor=executor), np.arange(12))


def test_arange_step(spec):
    a = xp.arange(20, step=3, chunks=(5,), spec=spec)
    assert_array_equal(a.compute(), np.arange(20, step=3))


def test_asarray(spec, executor):
    a = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 2), spec=spec)
    assert_array_equal(
        a.compute(executor=executor), np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    )


def test_asarray_from_array(spec):
    a = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 2), spec=spec)
    b = xp.asarray(a)
    assert b is a
    assert_array_equal(b.compute(), np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]))


@pytest.mark.parametrize("k", [-1, 0, 1])
def test_eye(spec, k):
    a = xp.eye(5, k=k, chunks=(2, 2), spec=spec)
    assert_array_equal(a, np.eye(5, k=k))


def test_eye_not_square():
    a = xp.eye(1, 3, k=2)
    assert_array_equal(a, np.eye(1, 3, k=2))


@pytest.mark.parametrize("endpoint", [True, False])
def test_linspace(spec, endpoint):
    a = xp.linspace(6, 49, 50, endpoint=endpoint, chunks=5, spec=spec)
    npa = np.linspace(6, 49, 50, endpoint=endpoint)
    assert_allclose(a, npa)

    a = xp.linspace(1.4, 4.9, 13, endpoint=endpoint, chunks=5, spec=spec)
    npa = np.linspace(1.4, 4.9, 13, endpoint=endpoint)
    assert_allclose(a, npa)

    a = xp.linspace(0, 0, 0, endpoint=endpoint)
    npa = np.linspace(0, 0, 0, endpoint=endpoint)
    assert_allclose(a, npa)


def test_ones(spec, executor):
    a = xp.ones((3, 3), chunks=(2, 2), spec=spec)
    assert_array_equal(a.compute(executor=executor), np.ones((3, 3)))


def test_ones_like(spec, executor):
    a = xp.ones((3, 3), chunks=(2, 2), spec=spec)
    b = xp.ones_like(a)
    assert_array_equal(b.compute(executor=executor), np.ones_like(np.ones((3, 3))))


@pytest.mark.parametrize("k", [-1, 0, 1])
def test_tril_triu(spec, k):
    a = xp.ones((4, 5), chunks=(2, 2), spec=spec)
    assert_array_equal(xp.tril(a, k=k), np.tril(np.ones((4, 5)), k))
    assert_array_equal(xp.triu(a, k=k), np.triu(np.ones((4, 5)), k))


# Data type functions


def test_astype(spec, executor):
    a = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 2), spec=spec)
    b = xp.astype(a, xp.int32)
    assert_array_equal(
        b.compute(executor=executor),
        np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]),
    )


# Elementwise functions


# test on all executors (`any_executor`) for coverage
def test_add(spec, any_executor):
    a = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 2), spec=spec)
    b = xp.asarray([[1, 1, 1], [1, 1, 1], [1, 1, 1]], chunks=(2, 2), spec=spec)
    c = xp.add(a, b)
    assert_array_equal(
        c.compute(executor=any_executor), np.array([[2, 3, 4], [5, 6, 7], [8, 9, 10]])
    )


def test_add_top_level_namespace(spec, executor):
    a = cubed.ones((10, 10), chunks=(10, 2), spec=spec)
    b = cubed.ones((10, 10), chunks=(2, 10), spec=spec)
    c = cubed.add(a, b)
    assert_array_equal(
        c.compute(executor=executor), np.ones((10, 10)) + np.ones((10, 10))
    )


def test_add_different_chunks(spec, executor):
    a = xp.ones((10, 10), chunks=(10, 2), spec=spec)
    b = xp.ones((10, 10), chunks=(2, 10), spec=spec)
    c = xp.add(a, b)
    assert_array_equal(
        c.compute(executor=executor), np.ones((10, 10)) + np.ones((10, 10))
    )


def test_add_different_chunks_fail(spec, executor):
    a = xp.ones((10,), chunks=(3,), spec=spec)
    b = xp.ones((10,), chunks=(5,), spec=spec)
    c = xp.add(a, b)
    assert_array_equal(c.compute(executor=executor), np.ones((10,)) + np.ones((10,)))


def test_add_scalars():
    a = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 2))

    b = xp.add(a, 1)
    assert_array_equal(b.compute(), np.array([[2, 3, 4], [5, 6, 7], [8, 9, 10]]))

    c = xp.add(2, a)
    assert_array_equal(c.compute(), np.array([[3, 4, 5], [6, 7, 8], [9, 10, 11]]))

    with pytest.raises(TypeError):
        xp.add(1, 2)


@pytest.mark.parametrize(
    ("min", "max"),
    [
        (None, None),
        (4, None),
        (None, 7),
        (4, 7),
        (0, 10),
    ],
)
def test_clip(spec, min, max):
    a = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 2), spec=spec)
    npa = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    b = xp.clip(a, min, max)
    if min is max is None:
        assert b is a
    else:
        assert_array_equal(b.compute(), np.clip(npa, min, max))


def test_equal(spec):
    a = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 2), spec=spec)
    b = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 2), spec=spec)
    c = xp.equal(a, b)
    assert_array_equal(c.compute(), np.full((3, 3), True))


def test_negative(spec, executor):
    a = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 2), spec=spec)
    b = xp.negative(a)
    assert_array_equal(
        b.compute(executor=executor),
        np.array([[-1, -2, -3], [-4, -5, -6], [-7, -8, -9]]),
    )


# Indexing


@pytest.mark.parametrize(
    "ind",
    [
        Ellipsis,
        6,
        (6, xp.newaxis),
        (xp.newaxis, 6),
        slice(None),
        slice(10),
        slice(3, None),
        slice(3, 3),
        slice(3, 10),
        (slice(10), xp.newaxis),
    ],
)
def test_index_1d(spec, ind):
    a = xp.arange(12, chunks=(4,), spec=spec)
    assert_array_equal(a[ind].compute(), np.arange(12)[ind])


@pytest.mark.parametrize(
    "ind",
    [
        Ellipsis,
        (2, 3),
        (xp.newaxis, 2, 3),
        (Ellipsis, slice(2, 4)),
        (slice(None), slice(2, 2)),
        (slice(None), slice(2, 4)),
        (slice(3), slice(2, None)),
        (slice(1, None), slice(4)),
        (slice(1, 3), Ellipsis),
        (slice(1, 1), slice(None)),
        (slice(1, 3), slice(None)),
        (xp.newaxis, slice(None), slice(2, 4)),
        (slice(None), xp.newaxis, slice(2, 4)),
        (slice(None), slice(2, 4), xp.newaxis),
        (slice(None), 1),
        (1, slice(2, 4)),
    ],
)
def test_index_2d(spec, ind):
    a = xp.asarray(
        [[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12], [13, 14, 15, 16]],
        chunks=(2, 2),
        spec=spec,
    )
    x = np.array([[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12], [13, 14, 15, 16]])
    assert_array_equal(a[ind].compute(), x[ind])


@pytest.mark.parametrize(
    "ind",
    [
        Ellipsis,
        (slice(None), slice(None)),
        (slice(0, 4), slice(None)),
        (slice(None), slice(0, 4)),
        (slice(0, 4), slice(0, 4)),
    ],
)
def test_index_2d_no_op(spec, ind):
    a = xp.asarray(
        [[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12], [13, 14, 15, 16]],
        chunks=(2, 2),
        spec=spec,
    )
    assert a is a[ind]


@pytest.mark.parametrize(
    ("shape", "chunks", "ind", "new_chunks_expected"),
    [
        # step divides chunks exactly
        (20, 4, slice(3, 14, 2), ((4, 2),)),
        # step doesn't divide chunks exactly
        # chunks is prime (so merge_chunks can't restore chunks to 5)
        (20, 5, slice(3, 14, 2), ((4, 2),)),
        # step doesn't divide chunks exactly
        (20, 8, slice(5, 18, 3), ((5,),)),
        # step is bigger than chunks
        (50, 5, slice(3, 50, 7), ((5, 2),)),
    ],
)
def test_index_1d_step(spec, shape, chunks, ind, new_chunks_expected):
    a = xp.arange(shape, chunks=chunks, spec=spec)
    b = a[ind]
    assert_array_equal(b.compute(), np.arange(shape)[ind])
    assert b.chunks == new_chunks_expected


# fmt: off
@pytest.mark.parametrize(
    ("shape", "chunks", "ind", "new_chunks_expected"),
    [
        (
            (20, 20),
            (4, 4),
            (slice(3, 14, 2), slice(3, 14, 3)),
            ((4, 2), (3, 1),),
        ),
    ],
)
# fmt: on
def test_index_2d_step(spec, shape, chunks, ind, new_chunks_expected):
    a = xp.ones(shape, chunks=chunks, spec=spec)
    b = a[ind]
    assert_array_equal(b.compute(), np.ones(shape)[ind])
    assert b.chunks == new_chunks_expected


@pytest.mark.parametrize(
    ("shape", "chunks", "ind"),
    [
        ((0, 2), (0, 2), (slice(None), 0)),
        ((0, 2), (1, 2), (slice(None), 0)),
        ((2, 0), (2, 0), (0, slice(None))),
        ((2, 0), (2, 1), (0, slice(None))),
    ],
)
def test_index_zero_dim(shape, chunks, ind):
    a = xp.ones(shape, chunks=chunks)
    b = a[ind]
    assert_array_equal(b.compute(), np.ones(shape)[ind])


def test_index_slice_unsupported_step(spec):
    a = xp.arange(12, chunks=(4,), spec=spec)
    with pytest.raises(NotImplementedError):
        a[::-1]


@pytest.mark.parametrize("axis", [0, 1])
@skip_if_cupy  # ndindex with a cupy.ndarray
def test_take(spec, axis):
    a = xp.asarray(
        [[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12], [13, 14, 15, 16]],
        chunks=(2, 2),
        spec=spec,
    )
    b = xp.asarray([1, 2], spec=spec)
    c = xp.take(a, b, axis=axis)
    x = np.array([[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12], [13, 14, 15, 16]])
    assert_array_equal(c.compute(), x.take([1, 2], axis=axis))


# Linear algebra functions


def test_matmul(spec, executor):
    a = xp.asarray(
        [[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12], [13, 14, 15, 16]],
        chunks=(2, 2),
        spec=spec,
    )
    b = xp.asarray(
        [[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12], [13, 14, 15, 16]],
        chunks=(2, 2),
        spec=spec,
    )
    c = xp.matmul(a, b)
    x = np.array([[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12], [13, 14, 15, 16]])
    y = np.array([[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12], [13, 14, 15, 16]])
    expected = np.matmul(x, y)
    assert_array_equal(c.compute(executor=executor), expected)


@pytest.mark.cloud
def test_matmul_cloud(executor):
    tmp_path = "s3://cubed-unittest/matmul"
    spec = cubed.Spec(tmp_path, allowed_mem=100000, storage_options=dict(use_obstore=True))

    a = xp.asarray(
        [[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12], [13, 14, 15, 16]],
        chunks=(2, 2),
        spec=spec,
    )
    b = xp.asarray(
        [[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12], [13, 14, 15, 16]],
        chunks=(2, 2),
        spec=spec,
    )
    c = xp.matmul(a, b)
    x = np.array([[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12], [13, 14, 15, 16]])
    y = np.array([[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12], [13, 14, 15, 16]])
    expected = np.matmul(x, y)
    assert_array_equal(c.compute(executor=executor), expected)


@pytest.mark.cloud
def test_matmul_modal(modal_executor):
    tmp_path = "s3://cubed-unittest/matmul"
    spec = cubed.Spec(tmp_path, allowed_mem=100000, storage_options=dict(use_obstore=True))

    a = xp.asarray(
        [[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12], [13, 14, 15, 16]],
        chunks=(2, 2),
        spec=spec,
    )
    b = xp.asarray(
        [[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12], [13, 14, 15, 16]],
        chunks=(2, 2),
        spec=spec,
    )
    c = xp.matmul(a, b)
    x = np.array([[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12], [13, 14, 15, 16]])
    y = np.array([[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12], [13, 14, 15, 16]])
    expected = np.matmul(x, y)
    assert_array_equal(c.compute(executor=modal_executor), expected)


def test_outer(spec, executor):
    a = xp.asarray([0, 1, 2], chunks=2, spec=spec)
    b = xp.asarray([10, 50, 100], chunks=2, spec=spec)
    c = xp.linalg.outer(a, b)
    assert_array_equal(c.compute(executor=executor), np.outer([0, 1, 2], [10, 50, 100]))


@pytest.mark.parametrize("axes", [1, (1, 0)])
def test_tensordot(axes):
    x = np.arange(400).reshape((20, 20))
    a = xp.asarray(x, chunks=(5, 4))
    y = np.arange(200).reshape((20, 10))
    b = xp.asarray(y, chunks=(4, 5))
    assert_array_equal(
        xp.tensordot(a, b, axes=axes).compute(), np.tensordot(x, y, axes=axes)
    )


# Manipulation functions


def test_broadcast_arrays(executor):
    a = xp.ones(30, chunks=(3,))
    b = xp.ones(30, chunks=(6,))
    a_b, b_b = xp.broadcast_arrays(a, b)

    assert_array_equal(a_b.compute(), np.ones(30))
    assert_array_equal(b_b.compute(), np.ones(30))

    a = xp.ones((1, 30), chunks=(1, 3))
    b = xp.ones(30, chunks=(6,))
    a_b, b_b = xp.broadcast_arrays(a, b)

    assert_array_equal(a_b.compute(executor=executor), np.ones((1, 30)))
    assert_array_equal(b_b.compute(executor=executor), np.ones((1, 30)))


@pytest.mark.parametrize(
    ("shape", "chunks", "new_shape", "new_chunks", "new_chunks_expected"),
    [
        ((), (), (0,), None, ((0,),)),
        ((5, 1, 6), (3, 1, 3), (5, 0, 6), None, ((3, 2), (0,), (3, 3))),
        ((5, 1, 6), (3, 1, 3), (5, 4, 6), None, ((3, 2), (1, 1, 1, 1), (3, 3))),
        ((5, 1, 6), (3, 1, 3), (2, 5, 1, 6), None, ((1, 1), (3, 2), (1,), (3, 3))),
        ((5, 1, 6), (3, 1, 3), (5, 3, 6), (3, 3, 3), ((3, 2), (3,), (3, 3))),
    ],
)
def test_broadcast_to(
    executor, shape, chunks, new_shape, new_chunks, new_chunks_expected
):
    x = np.random.randint(10, size=shape)
    a = xp.asarray(x, chunks=chunks)
    b = xp.broadcast_to(a, shape=new_shape, chunks=new_chunks)

    assert b.shape == new_shape
    assert b.chunks == new_chunks_expected
    assert_array_equal(b.compute(executor=executor), np.broadcast_to(x, new_shape))


def test_concat(spec, executor):
    # note: middle chunk of output reads from three input chunks
    a = xp.full((4, 5), 1, chunks=(3, 2), spec=spec)
    b = xp.full((1, 5), 2, chunks=(3, 2), spec=spec)
    c = xp.full((3, 5), 3, chunks=(3, 2), spec=spec)
    d = xp.concat([a, b, c], axis=0)
    assert_array_equal(
        d.compute(executor=executor),
        np.concatenate(
            [np.full((4, 5), 1), np.full((1, 5), 2), np.full((3, 5), 3)], axis=0
        ),
    )


def test_concat_different_chunks(spec):
    a = xp.asarray([[1], [5]], chunks=(2, 2), spec=spec)
    b = xp.asarray([[2, 3, 4], [6, 7, 8]], chunks=(2, 3), spec=spec)
    c = xp.concat([a, b], axis=1)
    assert_array_equal(
        c.compute(),
        np.concatenate(
            [
                np.array([[1], [5]]),
                np.array([[2, 3, 4], [6, 7, 8]]),
            ],
            axis=1,
        ),
    )


@pytest.mark.parametrize("axis", [None, 0])
def test_concat_single_array(spec, axis):
    a = xp.full((4, 5), 1, chunks=(3, 2), spec=spec)
    d = xp.concat([a], axis=axis)
    assert_array_equal(
        d.compute(),
        np.concatenate([np.full((4, 5), 1)], axis=axis),
    )


def test_concat_incompatible_shapes(spec):
    a = xp.full((4, 5), 1, chunks=(3, 2), spec=spec)
    b = xp.full((4, 6), 2, chunks=(3, 2), spec=spec)
    with pytest.raises(
        ValueError,
        match="all the input array dimensions except for the concatenation axis must match exactly",
    ):
        xp.concat([a, b], axis=0)
    xp.concat([a, b], axis=1)  # OK


def test_expand_dims(spec, executor):
    a = xp.asarray([1, 2, 3], chunks=(2,), spec=spec)
    b = xp.expand_dims(a, axis=0)
    assert_array_equal(b.compute(executor=executor), np.expand_dims([1, 2, 3], 0))


@pytest.mark.parametrize(
    ("shape", "chunks", "axis"),
    [
        ((10,), (4,), None),
        ((10,), (4,), 0),
        ((10,), (5,), 0),
        ((10, 7), (4, 3), None),
        ((10, 7), (4, 3), 0),
        ((10, 7), (4, 3), 1),
        ((10, 7), (4, 3), (0, 1)),
        ((10, 7), (4, 3), -1),
        ((10, 7), (5, 3), (0, 1)),
    ],
)
def test_flip(executor, shape, chunks, axis):
    x = np.random.randint(10, size=shape)
    a = xp.asarray(x, chunks=chunks)
    b = xp.flip(a, axis=axis)

    assert b.chunks == a.chunks

    assert_array_equal(
        b.compute(executor=executor),
        np.flip(x, axis=axis),
    )


def test_moveaxis(spec):
    a = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 2), spec=spec)
    b = xp.moveaxis(a, [0, -1], [-1, 0])
    assert_array_equal(
        b.compute(),
        np.moveaxis(np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]), [0, -1], [-1, 0]),
    )


def test_permute_dims(spec, executor):
    a = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 2), spec=spec)
    b = xp.permute_dims(a, (1, 0))
    assert_array_equal(
        b.compute(executor=executor),
        np.transpose(np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])),
    )


def test_repeat(spec):
    a = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 2), spec=spec)
    b = xp.repeat(a, 3, axis=1)
    assert_array_equal(
        b.compute(),
        np.repeat(np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]), 3, axis=1),
    )


def test_reshape(spec, executor):
    a = xp.arange(12, chunks=4, spec=spec)
    b = xp.reshape(a, (3, 4))

    assert_array_equal(
        b.compute(executor=executor),
        np.arange(12).reshape((3, 4)),
    )


def test_reshape_chunks(spec, executor):
    a = xp.arange(12, chunks=4, spec=spec)
    b = reshape_chunks(a, (2, 6), (2, 2))

    assert b.shape == (2, 6)
    assert b.chunks == ((2,), (2, 2, 2))

    assert_array_equal(
        b.compute(executor=executor),
        np.array([[0, 1, 4, 5, 8, 9], [2, 3, 6, 7, 10, 11]]),
    )


def test_reshape_chunks_with_smaller_end_chunk(spec, executor):
    a = xp.arange(10, chunks=4, spec=spec)
    b = reshape_chunks(a, (2, 5), (2, 2))

    assert b.shape == (2, 5)
    assert b.chunks == ((2,), (2, 2, 1))

    assert_array_equal(
        b.compute(executor=executor),
        np.array([[0, 1, 4, 5, 8], [2, 3, 6, 7, 9]]),
    )


def _maybe_len(a):
    try:
        return len(a)
    except TypeError:
        return 0


@pytest.mark.parametrize(
    ("chunks", "shift", "axis"),
    [
        ((2, 6), 3, None),
        ((2, 6), 3, 0),
        ((2, 6), (3, 9), (0, 1)),
        ((2, 6), (3, 9), None),
        ((2, 6), (3, 9), 1),
    ],
)
def test_roll(spec, executor, chunks, shift, axis):
    x = np.arange(4 * 6).reshape((4, 6))
    a = cubed.from_array(x, chunks=chunks, spec=spec)

    if _maybe_len(shift) != _maybe_len(axis):
        with pytest.raises(TypeError if axis is None else ValueError):
            xp.roll(a, shift, axis=axis)
    else:
        assert_array_equal(
            xp.roll(a, shift, axis=axis).compute(executor=executor),
            np.roll(x, shift, axis),
        )


def test_squeeze_1d(spec, executor):
    a = xp.asarray([[1, 2, 3]], chunks=(1, 2), spec=spec)
    b = xp.squeeze(a, 0)
    assert_array_equal(b.compute(executor=executor), np.squeeze([[1, 2, 3]], 0))


def test_squeeze_2d(spec, executor):
    a = xp.asarray([[[1], [2], [3]]], chunks=(1, 2, 1), spec=spec)
    b = xp.squeeze(a, (0, 2))
    assert_array_equal(
        b.compute(executor=executor), np.squeeze([[[1], [2], [3]]], (0, 2))
    )


def test_stack(spec, executor):
    a = xp.full((4, 6), 1, chunks=(2, 3), spec=spec)
    b = xp.full((4, 6), 2, chunks=(2, 3), spec=spec)
    c = xp.full((4, 6), 3, chunks=(2, 3), spec=spec)
    d = xp.stack([a, b, c], axis=0)
    assert_array_equal(
        d.compute(executor=executor),
        np.stack([np.full((4, 6), 1), np.full((4, 6), 2), np.full((4, 6), 3)], axis=0),
    )


@pytest.mark.parametrize("repetitions", [(2,), (2, 5), (2, 5, 3)])
def test_tile(spec, repetitions):
    a = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 2), spec=spec)
    b = xp.tile(a, repetitions)
    assert_array_equal(
        b.compute(),
        np.tile(np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]), repetitions),
    )


@pytest.mark.parametrize("chunks", [(1, 2, 3), (2, 2, 3), (3, 2, 3)])
def test_unstack(spec, executor, chunks):
    a = xp.full((4, 6), 1, chunks=(2, 3), spec=spec)
    b = xp.full((4, 6), 2, chunks=(2, 3), spec=spec)
    c = xp.full((4, 6), 3, chunks=(2, 3), spec=spec)
    d = xp.stack([a, b, c], axis=0)

    d = d.rechunk(chunks)

    au, bu, cu = cubed.compute(*xp.unstack(d), executor=executor, optimize_graph=False)

    assert_array_equal(au, np.full((4, 6), 1))
    assert_array_equal(bu, np.full((4, 6), 2))
    assert_array_equal(cu, np.full((4, 6), 3))


def test_unstack_zero_arrays(spec):
    a = xp.full((0, 4, 6), 1, chunks=(1, 2, 3), spec=spec)
    assert xp.unstack(a) == ()


def test_unstack_single_array(spec):
    a = xp.full((1, 4, 6), 1, chunks=(1, 2, 3), spec=spec)
    (b,) = xp.unstack(a)
    assert_array_equal(b.compute(), np.full((4, 6), 1))


# Searching functions


def test_argmax(spec):
    a = xp.asarray([[11, 12, 13], [11, 11, 14], [10, 13, 11]], chunks=(2, 2), spec=spec)
    b = xp.argmax(a)
    assert_array_equal(
        b.compute(),
        np.array([[11, 12, 13], [11, 11, 14], [10, 13, 11]]).argmax(),
    )


def test_argmax_axis_0(spec):
    a = xp.asarray([[11, 12, 13], [11, 11, 14], [10, 13, 11]], chunks=(2, 2), spec=spec)
    b = xp.argmax(a, axis=0)
    assert_array_equal(
        b.compute(),
        np.array([[11, 12, 13], [11, 11, 14], [10, 13, 11]]).argmax(axis=0),
    )


def test_argmin_axis_0(spec):
    a = xp.asarray([[11, 12, 13], [11, 11, 14], [10, 13, 11]], chunks=(2, 2), spec=spec)
    b = xp.argmin(a, axis=0)
    assert_array_equal(
        b.compute(),
        np.array([[11, 12, 13], [11, 11, 14], [10, 13, 11]]).argmin(axis=0),
    )


@pytest.mark.parametrize(
    ("x1", "x1_chunks", "x2", "x2_chunks"),
    [
        ([], 1, [], 1),
        ([0], 1, [0], 1),
        ([-10, 0, 10, 20, 30], 3, [11, 30], 2),
        ([-10, 0, 10, 20, 30], 3, [11, 30, -20, 1, -10, 10, 37, 11], 5),
        ([-10, 0, 10, 20, 30], 3, [[11, 30, -20, 1, -10, 10, 37, 11]], 5),
        ([-10, 0, 10, 20, 30], 3, [[7, 0], [-10, 10], [11, -1], [15, 15]], (2, 2)),
    ],
)
@pytest.mark.parametrize("side", ["left", "right"])
def test_searchsorted(x1, x1_chunks, x2, x2_chunks, side):
    x1 = np.array(x1)
    x2 = np.array(x2)

    x1d = xp.asarray(x1, chunks=x1_chunks)
    x2d = xp.asarray(x2, chunks=x2_chunks)

    out = xp.searchsorted(x1d, x2d, side=side)

    assert out.shape == x2d.shape
    assert out.chunks == x2d.chunks
    assert_array_equal(out.compute(), np.searchsorted(x1, x2, side=side))


def test_searchsorted_sorter_not_implemented():
    with pytest.raises(NotImplementedError):
        xp.searchsorted(xp.asarray([1, 0]), xp.asarray([1]), sorter=xp.asarray([1, 0]))


def test_where_scalars():
    condition = xp.asarray(
        [[True, False, True], [False, True, False], [True, False, True]], chunks=(2, 2)
    )
    a = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 2))

    b = xp.where(condition, a, 0)
    assert_array_equal(b.compute(), np.array([[1, 0, 3], [0, 5, 0], [7, 0, 9]]))

    c = xp.where(condition, 0, a)
    assert_array_equal(c.compute(), np.array([[0, 2, 0], [4, 0, 6], [0, 8, 0]]))

    with pytest.raises(TypeError):
        xp.where(condition, 0, 1)


# Statistical functions


@pytest.mark.parametrize("axis", [0, 1])
def test_cumulative_prod_2d(axis):
    a = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 2))
    b = xp.cumulative_prod(a, axis=axis)
    assert_array_equal(
        b.compute(),
        np.cumulative_prod(np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]), axis=axis),
    )


@pytest.mark.parametrize("axis", [0, 1])
def test_cumulative_sum_2d(axis):
    a = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 2))
    b = xp.cumulative_sum(a, axis=axis)
    assert_array_equal(
        b.compute(),
        np.cumulative_sum(np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]), axis=axis),
    )


def test_cumulative_sum_2d_recursive(executor):
    a = xp.ones((10, 100), chunks=(10, 10))
    b = xp.cumulative_sum(a, axis=1)
    assert_array_equal(
        b.compute(executor=executor),
        np.cumulative_sum(np.ones((10, 100)), axis=1),
    )


def test_cumulative_sum_1d():
    a = xp.asarray([0, 1, 2, 3, 4, 5, 6, 7, 8, 9], chunks=(4,))
    b = xp.cumulative_sum(a, axis=0)
    assert_array_equal(
        b.compute(),
        np.cumulative_sum(np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9]), axis=0),
    )


def test_cumulative_sum_unsupported_include_initial(spec):
    a = xp.asarray([0, 1, 2, 3, 4, 5, 6, 7, 8, 9], chunks=(4,))
    with pytest.raises(NotImplementedError):
        xp.cumulative_sum(a, axis=0, include_initial=True)


def test_mean_axis_0(spec, executor):
    a = xp.asarray(
        [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0]], chunks=(2, 2), spec=spec
    )
    b = xp.mean(a, axis=0)
    assert_array_equal(
        b.compute(executor=executor),
        np.array([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0]]).mean(axis=0),
    )

def test_mean_complex():
    a = xp.asarray([1.0+1.0j, 2.0+2.0j, 3.0+3.0j], chunks=(2,))
    b = xp.mean(a)
    assert_array_equal(b.compute(), np.array([1.0+1.0j, 2.0+2.0j, 3.0+3.0j]).mean())


def test_sum(spec, executor):
    a = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 2), spec=spec)
    b = xp.sum(a)
    assert_array_equal(
        b.compute(executor=executor), np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]).sum()
    )


def test_sum_axis_0(spec, executor):
    a = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 2), spec=spec)
    b = xp.sum(a, axis=0)
    assert_array_equal(b.compute(executor=executor), np.array([12, 15, 18]))


@pytest.mark.parametrize("axis", [None, 0, 1, (0, 1)])
@pytest.mark.parametrize("correction", [0.0, 1.0])
@pytest.mark.parametrize("keepdims", [False, True])
def test_var(spec, axis, correction, keepdims):
    a = xp.asarray(
        [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0]], chunks=(2, 2), spec=spec
    )
    b = xp.var(a, axis=axis, correction=correction, keepdims=keepdims)
    assert_array_equal(
        b.compute(optimize_graph=False),
        np.array([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0]]).var(
            axis=axis, ddof=correction, keepdims=keepdims
        ),
    )


@pytest.mark.parametrize("axis", [None, 0, 1, (0, 1)])
@pytest.mark.parametrize("correction", [0.0, 1.0])
@pytest.mark.parametrize("keepdims", [False, True])
def test_std(spec, axis, correction, keepdims):
    a = xp.asarray(
        [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0]], chunks=(2, 2), spec=spec
    )
    b = xp.std(a, axis=axis, correction=correction, keepdims=keepdims)
    assert_array_equal(
        b.compute(),
        np.array([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0]]).std(
            axis=axis, ddof=correction, keepdims=keepdims
        ),
    )


def test_var__poorly_conditioned(spec):
    # from https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance#Example
    npa = np.array([4.0, 7.0, 13.0, 16.0]) + 1e9
    a = xp.asarray(npa, chunks=2, spec=spec)
    b = xp.var(a, axis=0)
    assert_array_equal(b.compute(), npa.var(axis=0))


# Utility functions


def test_all(spec, executor):
    a = xp.asarray(
        [[True, True, True], [True, True, True], [True, True, True]],
        chunks=(2, 2),
        spec=spec,
    )
    b = xp.all(a)
    assert b.ndim == 0
    assert b.size == 1
    assert b.compute(executor=executor)


def test_all_zero_dimension(spec, executor):
    a = xp.ones((0,), spec=spec)
    b = xp.all(a)
    assert b.ndim == 0
    assert b.size == 1
    assert b.compute(executor=executor)


@pytest.mark.parametrize("n", [1, 2])
def test_diff(n):
    x = np.array([1, 5, 3, 8, 7, 2, 6, 9])
    a = xp.asarray(x, chunks=(3,))
    b = xp.diff(a, n=n)

    assert_array_equal(b.compute(), np.diff(x, n=n))


@pytest.mark.parametrize(
    ("shape", "axis"),
    [((10, 15, 20), 0), ((10, 15, 20), 1), ((10, 15, 20), 2), ((10, 15, 20), -1)],
)
@pytest.mark.parametrize("n", [0, 1, 2])
def test_diff_3d(shape, n, axis):
    x = np.random.default_rng().integers(0, 10, shape)
    a = xp.asarray(x, chunks=(len(shape) * (5,)))

    assert_array_equal(xp.diff(a, axis=axis, n=n), np.diff(x, axis=axis, n=n))



import platform
from functools import partial

import pytest

from cubed.tests.runtime.utils import check_invocation_counts, deterministic_failure

pytest.importorskip("lithops")

from lithops.executors import LocalhostExecutor
from lithops.retries import RetryingFunctionExecutor

from cubed.runtime.executors.lithops import map_unordered


def run_test(function, input, retries, timeout=10, use_backups=False):
    outputs = set()
    with RetryingFunctionExecutor(LocalhostExecutor()) as executor:
        for output in map_unordered(
            executor,
            [function],
            [input],
            ["group0"],
            timeout=timeout,
            retries=retries,
            use_backups=use_backups,
        ):
            outputs.add(output)
    return outputs


# fmt: off
@pytest.mark.parametrize(
    ("timing_map", "n_tasks", "retries"),
    [
        # no failures
        ({}, 3, 2),
        # first invocation fails
        ({0: [-1], 1: [-1], 2: [-1]}, 3, 2),
        # first two invocations fail
        ({0: [-1, -1], 1: [-1, -1], 2: [-1, -1]}, 3, 2),
        # first input sleeps once
        ({0: [20]}, 3, 2),
    ],
)
# fmt: on
@pytest.mark.parametrize("use_backups", [False, True])
def test_success(tmp_path, timing_map, n_tasks, retries, use_backups):
    outputs = run_test(
        function=partial(deterministic_failure, tmp_path, timing_map),
        input=range(n_tasks),
        retries=retries,
        use_backups=use_backups,
    )

    assert outputs == set(range(n_tasks))

    check_invocation_counts(tmp_path, timing_map, n_tasks, retries)


# fmt: off
@pytest.mark.parametrize(
    ("timing_map", "n_tasks", "retries"),
    [
        # too many failures
        ({0: [-1], 1: [-1], 2: [-1, -1, -1]}, 3, 2),
    ],
)
# fmt: on
@pytest.mark.parametrize("use_backups", [False, True])
def test_failure(tmp_path, timing_map, n_tasks, retries, use_backups):
    with pytest.raises(RuntimeError):
        run_test(
            function=partial(deterministic_failure, tmp_path, timing_map),
            input=range(n_tasks),
            retries=retries,
            use_backups=use_backups,
        )

    check_invocation_counts(tmp_path, timing_map, n_tasks, retries)


# fmt: off
@pytest.mark.parametrize(
    ("timing_map", "n_tasks", "retries"),
    [
        ({0: [60]}, 10, 2),
    ],
)
# fmt: on
@pytest.mark.skipif(platform.system() == "Windows", reason="does not run on windows")
def test_stragglers(tmp_path, timing_map, n_tasks, retries):
    # TODO: run a test like this using a cloud executor
    # Reason: this test passes, but lithops local mode only runs one job at a time,
    # so it actually waits for the first job to finish before running the second one.

    outputs = run_test(
        function=partial(deterministic_failure, tmp_path, timing_map),
        input=range(n_tasks),
        retries=retries,
        timeout=500,
        use_backups=True,
    )

    assert outputs == set(range(n_tasks))

    check_invocation_counts(tmp_path, timing_map, n_tasks, retries)



import builtins
import math
import numbers
from dataclasses import dataclass
from functools import partial
from itertools import product
from numbers import Integral, Number
from typing import TYPE_CHECKING, Any, Callable, Sequence, Tuple, Union
from warnings import warn

import numpy as np
import zarr
from tlz import concat, first, partition
from toolz import map

from cubed import config
from cubed.backend_array_api import IS_IMMUTABLE_ARRAY, numpy_array_to_backend_array
from cubed.backend_array_api import namespace as nxp
from cubed.core.array import CoreArray, check_array_specs, compute, gensym
from cubed.core.plan import Plan, context_dir_path
from cubed.primitive.blockwise import blockwise as primitive_blockwise
from cubed.primitive.blockwise import general_blockwise as primitive_general_blockwise
from cubed.primitive.memory import get_buffer_copies
from cubed.primitive.rechunk import rechunk as primitive_rechunk
from cubed.spec import spec_from_config
from cubed.storage.backend import is_backend_storage_array, open_backend_array
from cubed.storage.zarr import lazy_zarr_array
from cubed.types import T_RegularChunks, T_Shape
from cubed.utils import (
    array_memory,
    array_size,
    get_item,
    itemsize,
    normalize_chunks,
    offset_to_block_id,
    to_chunksize,
)
from cubed.utils import numblocks as compute_numblocks
from cubed.vendor.dask.array.utils import validate_axis
from cubed.vendor.dask.blockwise import broadcast_dimensions
from cubed.vendor.dask.utils import has_keyword
from cubed.vendor.rechunker.algorithm import multistage_rechunking_plan

if TYPE_CHECKING:
    from cubed.array_api.array_object import Array


def from_array(x, chunks="auto", asarray=None, spec=None) -> "Array":
    """Create a Cubed array from an array-like object."""

    if isinstance(x, CoreArray):
        raise ValueError(
            "Array is already a Cubed array. Use 'asarray' or 'rechunk' instead."
        )

    previous_chunks = getattr(x, "chunks", None)
    outchunks = normalize_chunks(
        chunks, x.shape, dtype=x.dtype, previous_chunks=previous_chunks
    )

    if isinstance(x, zarr.Array):  # zarr fast path
        from cubed.array_api import Array

        name = gensym()
        target = x
        plan = Plan._new(name, "from_array", target)
        arr = Array(name, target, spec, plan)

        chunksize = to_chunksize(outchunks)
        if chunks != "auto" and previous_chunks != chunksize:
            arr = rechunk(arr, chunksize)
        return arr

    if asarray is None:
        asarray = not hasattr(x, "__array_function__")

    return map_blocks(
        _from_array,
        dtype=x.dtype,
        chunks=outchunks,
        spec=spec,
        input_array=x,
        outchunks=outchunks,
        asarray=asarray,
    )


def _from_array(block, input_array, outchunks=None, asarray=None, block_id=None):
    out = input_array[get_item(outchunks, block_id)]
    if asarray:
        out = np.asarray(out)
    out = numpy_array_to_backend_array(out)
    return out


def from_zarr(store, path=None, spec=None) -> "Array":
    """Load an array from Zarr storage.

    Parameters
    ----------
    store : string or Zarr Store
        Input Zarr store
    path : string, optional
        Group path
    spec : cubed.Spec, optional
        The spec to use for the computation.

    Returns
    -------
    cubed.Array
        The array loaded from Zarr storage.
    """
    name = gensym()
    spec = spec or spec_from_config(config)
    target = open_backend_array(
        store,
        mode="r",
        path=path,
        storage_options=spec.storage_options,
    )

    from cubed.array_api import Array

    plan = Plan._new(name, "from_zarr", target)
    return Array(name, target, spec, plan)


def store(
    sources: Union["Array", Sequence["Array"]],
    targets,
    regions: tuple[slice, ...] | list[tuple[slice, ...]] | None = None,
    executor=None,
    **kwargs,
):
    """Save source arrays to array-like objects.

    In the current implementation ``targets`` must be Zarr arrays.

    Note that this operation is eager, and will run the computation
    immediately.

    Parameters
    ----------
    sources : cubed.Array or collection of cubed.Array
        Arrays to save
    targets : string or Zarr store or collection of strings or Zarr stores
        Zarr arrays to write to
    regions : tuple of slices or list of tuple of slices, optional
        The regions of data that should be written to in targets.
    executor : cubed.runtime.types.Executor, optional
        The executor to use to run the computation.
        Defaults to using the in-process Python executor.
    """
    if isinstance(sources, CoreArray):
        sources = [sources]
        targets = [targets]

    if any(not isinstance(s, CoreArray) for s in sources):
        raise ValueError("All sources must be cubed array objects")

    if len(sources) != len(targets):
        raise ValueError(
            f"Different number of sources ({len(sources)}) and targets ({len(targets)})"
        )

    if isinstance(regions, tuple) or regions is None:
        regions_list = [regions] * len(sources)
    else:
        regions_list = list(regions)
        if len(sources) != len(regions_list):
            raise ValueError(
                f"Different number of sources [{len(sources)}] and "
                f"targets [{len(targets)}] than regions [{len(regions_list)}]"
            )

    arrays = []
    for source, target, region in zip(sources, targets, regions_list):
        array = _store_array(source, target, region=region)
        arrays.append(array)
    compute(*arrays, executor=executor, _return_in_memory_array=False, **kwargs)


def _store_array(source: "Array", target, path=None, region=None):
    if target is not None and not is_backend_storage_array(target):
        target = lazy_zarr_array(
            target,
            shape=source.shape,
            dtype=source.dtype,
            chunks=source.chunksize,
            path=path,
        )
    if target is None and region is not None:
        raise ValueError("Target store must be specified when setting a region")
    identity = lambda a: a
    if region is None or all(r == slice(None) for r in region):
        ind = tuple(range(source.ndim))
        return blockwise(
            identity,
            ind,
            source,
            ind,
            dtype=source.dtype,
            align_arrays=False,
            target_store=target,
        )
    else:
        # treat a region as an offset within the target store
        shape = target.shape
        chunks = target.chunks
        for i, (sl, cs) in enumerate(zip(region, chunks)):
            if sl.start % cs != 0 or (sl.stop % cs != 0 and sl.stop != shape[i]):
                raise ValueError(
                    f"Region {region} does not align with target chunks {chunks}"
                )
        block_offsets = [sl.start // cs for sl, cs in zip(region, chunks)]

        def key_function(out_key):
            out_coords = out_key[1:]
            in_coords = tuple(bi - off for bi, off in zip(out_coords, block_offsets))
            return ((source.name, *in_coords),)

        # calculate output block ids from region selection
        indexer = _create_zarr_indexer(region, shape, chunks)
        if source.shape != indexer.shape:
            raise ValueError(
                f"Source array shape {source.shape} does not match region shape {indexer.shape}"
            )
        # TODO(#800): make Zarr indexer pickle-able so we don't have to materialize all the block IDs
        output_blocks = map(
            lambda chunk_projection: list(chunk_projection[0]), list(indexer)
        )

        out = general_blockwise(
            identity,
            key_function,
            source,
            shapes=[shape],
            dtypes=[source.dtype],
            chunkss=[chunks],
            target_stores=[target],
            output_blocks=output_blocks,
        )
        from cubed import Array

        assert isinstance(out, Array)  # single output
        return out


def to_zarr(x: "Array", store, path=None, region=None, executor=None, **kwargs):
    """Save an array to Zarr storage.

    Note that this operation is eager, and will run the computation
    immediately.

    Parameters
    ----------
    x : cubed.Array
        Array to save
    store : string or Zarr store
        Output Zarr store
    path : string, optional
        Group path
    region : tuple of slices, optional
        The region of data that should be written to in target.
    executor : cubed.runtime.types.Executor, optional
        The executor to use to run the computation.
        Defaults to using the in-process Python executor.
    """
    out = _store_array(x, store, path=path, region=region)
    out.compute(executor=executor, _return_in_memory_array=False, **kwargs)


def blockwise(
    func,
    out_ind,
    *args: Any,  # can't type this as mypy assumes args are all same type, but blockwise args alternate types
    dtype=None,
    adjust_chunks=None,
    new_axes=None,
    align_arrays=True,
    target_store=None,
    target_path=None,
    extra_func_kwargs=None,
    **kwargs,
) -> "Array":
    arrays = args[::2]

    assert len(arrays) > 0

    # Calculate output chunking. A lot of this is from dask's blockwise function

    new_axes = new_axes or {}
    new = (
        set(out_ind)
        - {a for arg in args[1::2] if arg is not None for a in arg}
        - set(new_axes or ())
    )
    if new:
        raise ValueError("Unknown dimension", new)

    if align_arrays:
        chunkss, arrays = unify_chunks(*args)
    else:
        inds = args[1::2]
        arginds = zip(arrays, inds)

        chunkss = {}
        # For each dimension, use the input chunking that has the most blocks;
        # this will ensure that broadcasting works as expected, and in
        # particular the number of blocks should be correct if the inputs are
        # consistent.
        for arg, ind in arginds:
            arg_chunks = normalize_chunks(
                arg.chunks, shape=arg.shape, dtype=arg.dtype
            )  # have to normalize zarr chunks
            for c, i in zip(arg_chunks, ind):
                if i not in chunkss or len(c) > len(chunkss[i]):
                    chunkss[i] = c

    for k, v in new_axes.items():
        if not isinstance(v, tuple):
            v = (v,)
        chunkss[k] = v

    chunks = [chunkss[i] for i in out_ind]
    if adjust_chunks:
        for i, ind in enumerate(out_ind):
            if ind in adjust_chunks:
                if callable(adjust_chunks[ind]):
                    chunks[i] = tuple(map(adjust_chunks[ind], chunks[i]))
                elif isinstance(adjust_chunks[ind], numbers.Integral):
                    chunks[i] = tuple(adjust_chunks[ind] for _ in chunks[i])
                elif isinstance(adjust_chunks[ind], (tuple, list)):
                    if len(adjust_chunks[ind]) != len(chunks[i]):
                        raise ValueError(
                            f"Dimension {i} has {len(chunks[i])} blocks, adjust_chunks "
                            f"specified with {len(adjust_chunks[ind])} blocks"
                        )
                    chunks[i] = tuple(adjust_chunks[ind])
                else:
                    raise NotImplementedError(
                        "adjust_chunks values must be callable, int, or tuple"
                    )
    _chunks = tuple(chunks)
    shape = tuple(map(sum, _chunks))

    # replace arrays with zarr arrays
    zargs = list(args)
    zargs[::2] = [a._zarray for a in arrays]
    in_names = [a.name for a in arrays]

    extra_source_arrays = kwargs.pop("extra_source_arrays", [])
    source_arrays = list(arrays) + list(extra_source_arrays)

    extra_projected_mem = kwargs.pop("extra_projected_mem", 0)

    fusable_with_predecessors = kwargs.pop("fusable_with_predecessors", True)
    fusable_with_successors = kwargs.pop("fusable_with_successors", True)
    num_input_blocks = kwargs.pop("num_input_blocks", None)
    iterable_input_blocks = kwargs.pop("iterable_input_blocks", None)

    name = gensym()
    spec = check_array_specs(arrays)
    buffer_copies = get_buffer_copies(spec)
    if target_store is None:
        target_store = context_dir_path(spec=spec)
    op = primitive_blockwise(
        func,
        out_ind,
        *zargs,
        allowed_mem=spec.allowed_mem,
        reserved_mem=spec.reserved_mem,
        extra_projected_mem=extra_projected_mem,
        target_store=target_store,
        target_name=name,
        target_path=target_path,
        storage_options=spec.storage_options,
        compressor=spec.zarr_compressor,
        shape=shape,
        dtype=dtype,
        chunks=_chunks,
        new_axes=new_axes,
        in_names=in_names,
        buffer_copies=buffer_copies,
        extra_func_kwargs=extra_func_kwargs,
        fusable_with_predecessors=fusable_with_predecessors,
        fusable_with_successors=fusable_with_successors,
        num_input_blocks=num_input_blocks,
        iterable_input_blocks=iterable_input_blocks,
        **kwargs,
    )
    plan = Plan._new(
        name,
        "blockwise",
        op.target_array,
        op,
        False,
        *source_arrays,
    )
    from cubed.array_api import Array

    return Array(name, op.target_array, spec, plan)


def general_blockwise(
    func,
    key_function,
    *arrays,
    shapes,
    dtypes,
    chunkss,
    target_stores=None,
    target_paths=None,
    extra_func_kwargs=None,
    **kwargs,
) -> Union["Array", Tuple["Array", ...]]:
    if has_keyword(func, "block_id"):
        from cubed.array_api.creation_functions import offsets_virtual_array

        # Create an array of index offsets with the same chunk structure as the args,
        # which we convert to block ids (chunk coordinates) later.
        array0 = arrays[0]
        # note that primitive general_blockwise checks that all chunkss have same numblocks
        numblocks = compute_numblocks(chunkss[0])
        offsets = offsets_virtual_array(numblocks, array0.spec)
        new_arrays = arrays + (offsets,)

        def key_function_with_offset(key_function):
            def wrap(out_key):
                out_coords = out_key[1:]
                offset_in_key = ((offsets.name,) + out_coords,)
                return key_function(out_key) + offset_in_key

            return wrap

        def func_with_block_id(func):
            def wrap(*a, **kw):
                offset = int(a[-1])  # convert from 0-d array
                block_id = offset_to_block_id(offset, numblocks)
                return func(*a[:-1], block_id=block_id, **kw)

            return wrap

        function_nargs = kwargs.pop("function_nargs", None)
        if function_nargs is not None:
            function_nargs = function_nargs + 1  # for offsets array
        num_input_blocks = kwargs.pop("num_input_blocks", None)
        if num_input_blocks is not None:
            num_input_blocks = num_input_blocks + (1,)  # for offsets array
        iterable_input_blocks = kwargs.pop("iterable_input_blocks", None)
        if iterable_input_blocks is not None:
            iterable_input_blocks = iterable_input_blocks + (
                False,
            )  # for offsets array

        return _general_blockwise(
            func_with_block_id(func),
            key_function_with_offset(key_function),
            *new_arrays,
            shapes=shapes,
            dtypes=dtypes,
            chunkss=chunkss,
            target_stores=target_stores,
            target_paths=target_paths,
            extra_func_kwargs=extra_func_kwargs,
            function_nargs=function_nargs,
            num_input_blocks=num_input_blocks,
            iterable_input_blocks=iterable_input_blocks,
            **kwargs,
        )

    return _general_blockwise(
        func,
        key_function,
        *arrays,
        shapes=shapes,
        dtypes=dtypes,
        chunkss=chunkss,
        target_stores=target_stores,
        target_paths=target_paths,
        extra_func_kwargs=extra_func_kwargs,
        **kwargs,
    )


def _general_blockwise(
    func,
    key_function,
    *arrays,
    shapes,
    dtypes,
    chunkss,
    target_stores=None,
    target_paths=None,
    extra_func_kwargs=None,
    **kwargs,
) -> Union["Array", Tuple["Array", ...]]:
    assert len(arrays) > 0

    # replace arrays with zarr arrays
    zargs = [a._zarray for a in arrays]
    in_names = [a.name for a in arrays]

    extra_source_arrays = kwargs.pop("extra_source_arrays", [])
    source_arrays = list(arrays) + list(extra_source_arrays)

    extra_projected_mem = kwargs.pop("extra_projected_mem", 0)

    num_input_blocks = kwargs.pop("num_input_blocks", None)
    iterable_input_blocks = kwargs.pop("iterable_input_blocks", None)

    op_name = kwargs.pop("op_name", "blockwise")

    spec = check_array_specs(arrays)
    buffer_copies = get_buffer_copies(spec)

    if isinstance(target_stores, list) and len(target_stores) > 1:  # multiple outputs
        name = [gensym() for _ in range(len(target_stores))]
        target_stores = [
            ts if ts is not None else context_dir_path(spec=spec)
            for ts in target_stores
        ]
        target_names = name
    else:  # single output
        name = gensym()
        if target_stores is None:
            target_stores = [context_dir_path(spec=spec)]
        target_names = [name]

    op = primitive_general_blockwise(
        func,
        key_function,
        *zargs,
        allowed_mem=spec.allowed_mem,
        reserved_mem=spec.reserved_mem,
        extra_projected_mem=extra_projected_mem,
        buffer_copies=buffer_copies,
        target_stores=target_stores,
        target_names=target_names,
        target_paths=target_paths,
        storage_options=spec.storage_options,
        compressor=spec.zarr_compressor,
        shapes=shapes,
        dtypes=dtypes,
        chunkss=chunkss,
        in_names=in_names,
        extra_func_kwargs=extra_func_kwargs,
        num_input_blocks=num_input_blocks,
        iterable_input_blocks=iterable_input_blocks,
        **kwargs,
    )
    plan = Plan._new(
        name,
        op_name,
        op.target_array,
        op,
        False,
        *source_arrays,
    )
    from cubed.array_api import Array

    if isinstance(op.target_array, list):  # multiple outputs
        return tuple(Array(n, ta, spec, plan) for n, ta in zip(name, op.target_array))
    else:  # single output
        return Array(name, op.target_array, spec, plan)


def elemwise(func, *args: "Array", dtype=None) -> "Array":
    """Apply a function elementwise to array arguments, respecting broadcasting."""
    shapes = [arg.shape for arg in args]
    out_ndim = len(np.broadcast_shapes(*shapes))
    expr_inds = tuple(range(out_ndim))[::-1]
    if dtype is None:
        raise ValueError("dtype must be specified for elemwise")
    return blockwise(
        func,
        expr_inds,
        *concat((a, tuple(range(a.ndim)[::-1])) for a in args),
        dtype=dtype,
    )


def _create_zarr_indexer(selection, shape, chunks):
    if zarr.__version__[0] == "3":
        from zarr.core.chunk_grids import RegularChunkGrid
        from zarr.core.indexing import OrthogonalIndexer

        return OrthogonalIndexer(selection, shape, RegularChunkGrid(chunk_shape=chunks))
    else:
        from zarr.indexing import OrthogonalIndexer

        return OrthogonalIndexer(selection, ZarrArrayIndexingAdaptor(shape, chunks))


@dataclass
class ZarrArrayIndexingAdaptor:
    _shape: T_Shape
    _chunks: T_RegularChunks

    @classmethod
    def from_zarr_array(cls, zarray):
        return cls(zarray.shape, zarray.chunks)


def _assemble_index_chunk(
    arrays,
    dtype=None,
    func=None,
    selection_function=None,
    in_shape=None,
    in_chunksize=None,
    block_id=None,
    **kwargs,
):
    assert not isinstance(arrays, list), (
        "index expects an iterator of array blocks, not a list"
    )

    # compute the selection on x required to get the relevant chunk for out_coords
    out_coords = block_id
    in_sel = selection_function(("out",) + out_coords)

    # use a Zarr indexer to convert this to input coordinates
    indexer = _create_zarr_indexer(in_sel, in_shape, in_chunksize)

    shape = indexer.shape
    out = nxp.empty(shape, dtype=dtype)

    if array_size(shape) > 0:
        _, lchunk_selection, lout_selection, *_ = zip(*indexer)
        for ai, chunk_select, out_select in zip(
            arrays, lchunk_selection, lout_selection
        ):
            if IS_IMMUTABLE_ARRAY:
                out = out.at[out_select].set(ai[chunk_select])
            else:
                out[out_select] = ai[chunk_select]

    if func is not None:
        if has_keyword(func, "block_id"):
            out = func(out, block_id=block_id, **kwargs)
        else:
            out = func(out, **kwargs)
    return out


def map_selection(
    func,
    selection_function,
    x,
    shape,
    dtype,
    chunks,
    max_num_input_blocks,
    **kwargs,
) -> "Array":
    """
    Apply a function to selected subsets of an input array using standard NumPy indexing notation.

    Parameters
    ----------
    func : callable
        Function to apply to every block to produce the output array.
        Must accept ``block_id`` as a keyword argument (with same meaning as for ``map_blocks``).
    selection_function : callable
        A function that maps an output chunk key to one or more selections on the input array.
    x: Array
        The input array.
    shape : tuple
        Shape of the output array.
    dtype : np.dtype
        The ``dtype`` of the output array.
    chunks : tuple
        Chunk shape of blocks in the output array.
    max_num_input_blocks : int
        The maximum number of input blocks read from the input array.
    """

    def key_function(out_key):
        # compute the selection on x required to get the relevant chunk for out_key
        in_sel = selection_function(out_key)

        # use a Zarr indexer to convert selection to input coordinates
        indexer = _create_zarr_indexer(in_sel, x.shape, x.chunksize)

        return (iter(tuple((x.name,) + cp.chunk_coords for cp in indexer)),)

    num_input_blocks = (max_num_input_blocks,)
    iterable_input_blocks = (True,)

    out = general_blockwise(
        _assemble_index_chunk,
        key_function,
        x,
        shapes=[shape],
        dtypes=[dtype],
        chunkss=[chunks],
        extra_func_kwargs=dict(func=func, dtype=x.dtype),
        num_input_blocks=num_input_blocks,
        iterable_input_blocks=iterable_input_blocks,
        selection_function=selection_function,
        in_shape=x.shape,
        in_chunksize=x.chunksize,
        **kwargs,
    )
    from cubed import Array

    assert isinstance(out, Array)  # single output
    return out


def map_blocks(
    func,
    *args: "Array",
    dtype=None,
    chunks=None,
    drop_axis=None,
    new_axis=None,
    spec=None,
    **kwargs,
) -> "Array":
    """Apply a function to corresponding blocks from multiple input arrays."""

    from cubed.array_api.creation_functions import asarray

    # Coerce all args to Cubed arrays
    specs = [a.spec for a in args if hasattr(a, "spec")]
    spec0 = specs[0] if len(specs) > 0 else spec
    args = tuple(asarray(a, spec=spec0) for a in args)

    # Handle the case where an array is created by calling `map_blocks` with no input arrays
    if len(args) == 0:
        from cubed.array_api.creation_functions import empty_virtual_array

        shape = tuple(map(sum, chunks))
        args = (empty_virtual_array(shape, dtype=dtype, chunks=chunks, spec=spec),)

    if has_keyword(func, "block_id"):
        from cubed.array_api.creation_functions import offsets_virtual_array

        # Create an array of index offsets with the same chunk structure as the args,
        # which we convert to block ids (chunk coordinates) later.
        arg0 = args[0]
        numblocks = arg0.numblocks
        offsets = offsets_virtual_array(numblocks, arg0.spec)
        new_args = args + (offsets,)

        def func_with_block_id(func):
            def wrap(*a, **kw):
                offset = int(a[-1])  # convert from 0-d array
                block_id = offset_to_block_id(offset, numblocks)
                return func(*a[:-1], block_id=block_id, **kw)

            return wrap

        return _map_blocks(
            func_with_block_id(func),
            *new_args,
            dtype=dtype,
            chunks=chunks,
            drop_axis=drop_axis,
            new_axis=new_axis,
            **kwargs,
        )

    return _map_blocks(
        func,
        *args,
        dtype=dtype,
        chunks=chunks,
        drop_axis=drop_axis,
        new_axis=new_axis,
        **kwargs,
    )


def _map_blocks(
    func,
    *args: "Array",
    dtype=None,
    chunks=None,
    drop_axis=None,
    new_axis=None,
    **kwargs,
) -> "Array":
    # based on dask

    if drop_axis is None:
        drop_axis = []

    new_axes = {}

    if isinstance(drop_axis, Number):
        drop_axis = [drop_axis]
    if isinstance(new_axis, Number):
        new_axis = [new_axis]

    arrs = args
    argpairs = [
        (a, tuple(range(a.ndim))[::-1]) if isinstance(a, CoreArray) else (a, None)
        for a in args
    ]
    if arrs:
        out_ind = tuple(range(max(a.ndim for a in arrs)))[::-1]
    else:
        out_ind = ()

    if drop_axis:
        ndim_out = len(out_ind)
        if any(i < -ndim_out or i >= ndim_out for i in drop_axis):
            raise ValueError(
                f"drop_axis out of range (drop_axis={drop_axis}, "
                f"but output is {ndim_out}d)."
            )
        drop_axis = [i % ndim_out for i in drop_axis]
        out_ind = tuple(x for i, x in enumerate(out_ind) if i not in drop_axis)
    if new_axis is None and chunks is not None and len(out_ind) < len(chunks):
        new_axis = range(len(chunks) - len(out_ind))
    if new_axis:
        # new_axis = [x + len(drop_axis) for x in new_axis]
        temp_out_ind = list(out_ind)
        for ax in sorted(new_axis):
            n = len(temp_out_ind) + len(drop_axis)
            temp_out_ind.insert(ax, n)
            if chunks is not None:
                new_axes[n] = chunks[ax]
            else:
                new_axes[n] = 1
        out_ind = tuple(temp_out_ind)
        if max(new_axis) > max(out_ind):
            raise ValueError("New_axis values do not fill in all dimensions")

    if chunks is not None:
        if len(chunks) != len(out_ind):
            raise ValueError(
                f"Provided chunks have {len(chunks)} dims; expected {len(out_ind)} dims"
            )
        adjust_chunks = dict(zip(out_ind, chunks))
    else:
        adjust_chunks = None

    return blockwise(
        func,
        out_ind,
        *concat(argpairs),
        dtype=dtype,
        adjust_chunks=adjust_chunks,
        new_axes=new_axes,
        align_arrays=False,
        **kwargs,
    )


def map_direct(
    func, *args: "Array", shape, dtype, chunks, extra_projected_mem, spec=None, **kwargs
) -> "Array":
    """
    Apply a function across blocks of a new array, reading directly from side inputs (not necessarily in a blockwise fashion).

    Parameters
    ----------
    func : callable
        Function to apply to every block to produce the output array.
        Must accept ``block_id`` as a keyword argument (with same meaning as for ``map_blocks``).
    args : arrays
        The side-input arrays that may be accessed directly in the function.
    shape : tuple
        Shape of the output array.
    dtype : np.dtype
        The ``dtype`` of the output array.
    chunks : tuple
        Chunk shape of blocks in the output array.
    extra_projected_mem : int
        Extra memory projected to be needed (in bytes) for each map task. This should take into account the
        memory allocations for any reads from the side-input arrays (``args``).
    spec : Spec
        Specification for the new array. If not specified, the one from the first side input
        (`args`) will be used (if any).
    """
    warn(
        "`map_direct` is pending deprecation, please use `map_selection` instead",
        PendingDeprecationWarning,
        stacklevel=2,
    )

    from cubed.array_api.creation_functions import empty_virtual_array

    if spec is None and len(args) > 0 and hasattr(args[0], "spec"):
        spec = args[0].spec

    out = empty_virtual_array(shape, dtype=dtype, chunks=chunks, spec=spec)

    kwargs["arrays"] = args

    def new_func(func):
        def wrap(*a, block_id=None, **kw):
            arrays = kw.pop("arrays")
            args = a + arrays
            return func(*args, block_id=block_id, **kw)

        return wrap

    return map_blocks(
        new_func(func),
        out,
        dtype=dtype,
        chunks=chunks,
        extra_source_arrays=args,
        extra_projected_mem=extra_projected_mem,
        fusable_with_predecessors=False,  # don't allow fusion with predecessors since side inputs are not accounted for
        **kwargs,
    )


def rechunk(x, chunks, *, target_store=None, min_mem=None, use_new_impl=True):
    """Change the chunking of an array without changing its shape or data.

    Parameters
    ----------
    chunks : tuple
        The desired chunks of the array after rechunking.

    Returns
    -------
    cubed.Array
        An array with the desired chunks.
    """
    if use_new_impl:
        return rechunk_new(x, chunks, min_mem=min_mem)

    if isinstance(chunks, dict):
        chunks = {validate_axis(c, x.ndim): v for c, v in chunks.items()}
        for i in range(x.ndim):
            if i not in chunks:
                chunks[i] = x.chunks[i]
            elif chunks[i] is None:
                chunks[i] = x.chunks[i]
    if isinstance(chunks, (tuple, list)):
        chunks = tuple(lc if lc is not None else rc for lc, rc in zip(chunks, x.chunks))

    normalized_chunks = normalize_chunks(chunks, x.shape, dtype=x.dtype)
    if x.chunks == normalized_chunks:
        return x
    # normalizing takes care of dict args for chunks
    target_chunks = to_chunksize(normalized_chunks)

    # merge chunks special case
    if all(c1 % c0 == 0 for c0, c1 in zip(x.chunksize, target_chunks)):
        return merge_chunks(x, target_chunks)

    name = gensym()
    spec = x.spec
    if target_store is None:
        target_store = context_dir_path(spec=spec)
    name_int = f"{name}-int"
    ops = primitive_rechunk(
        x._zarray,
        source_array_name=x.name,
        int_array_name=name_int,
        target_array_name=name,
        target_chunks=target_chunks,
        allowed_mem=spec.allowed_mem,
        reserved_mem=spec.reserved_mem,
        target_store=target_store,
        storage_options=spec.storage_options,
    )

    from cubed.array_api import Array

    if len(ops) == 1:
        op = ops[0]
        plan = Plan._new(
            name,
            "rechunk",
            op.target_array,
            op,
            False,
            x,
        )
        return Array(name, op.target_array, spec, plan)

    else:
        op1 = ops[0]
        plan1 = Plan._new(
            name_int,
            "rechunk",
            op1.target_array,
            op1,
            False,
            x,
        )
        x_int = Array(name_int, op1.target_array, spec, plan1)

        op2 = ops[1]
        plan2 = Plan._new(
            name,
            "rechunk",
            op2.target_array,
            op2,
            False,
            x_int,
        )
        return Array(name, op2.target_array, spec, plan2)


def rechunk_new(x, chunks, *, min_mem=None):
    """Change the chunking of an array without changing its shape or data.

    Parameters
    ----------
    chunks : tuple
        The desired chunks of the array after rechunking.

    Returns
    -------
    cubed.Array
        An array with the desired chunks.
    """
    out = x
    for copy_chunks, target_chunks in _rechunk_plan(x, chunks, min_mem=min_mem):
        out = _rechunk(out, copy_chunks, target_chunks)
    return out


def _rechunk_plan(x, chunks, *, min_mem=None):
    if isinstance(chunks, dict):
        chunks = {validate_axis(c, x.ndim): v for c, v in chunks.items()}
        for i in range(x.ndim):
            if i not in chunks:
                chunks[i] = x.chunks[i]
            elif chunks[i] is None:
                chunks[i] = x.chunks[i]
    if isinstance(chunks, (tuple, list)):
        chunks = tuple(lc if lc is not None else rc for lc, rc in zip(chunks, x.chunks))

    normalized_chunks = normalize_chunks(chunks, x.shape, dtype=x.dtype)
    if x.chunks == normalized_chunks:
        return
    # normalizing takes care of dict args for chunks
    target_chunks = to_chunksize(normalized_chunks)

    spec = x.spec
    source_chunks = to_chunksize(normalize_chunks(x.chunks, x.shape, dtype=x.dtype))

    # rechunker doesn't take account of uncompressed and compressed copies of the
    # input and output array chunk/selection, so adjust appropriately:
    #  1 input array plus copies to read that array from storage,
    #  1 array for processing,
    #  1 output array plus copies to write that array to storage
    buffer_copies = get_buffer_copies(spec)
    total_copies = 1 + buffer_copies.read + 1 + 1 + buffer_copies.write
    rechunker_max_mem = (spec.allowed_mem - spec.reserved_mem) // total_copies
    if min_mem is None:
        min_mem = min(rechunker_max_mem // 20, x.nbytes)
    stages = multistage_rechunking_plan(
        shape=x.shape,
        source_chunks=source_chunks,
        target_chunks=target_chunks,
        itemsize=itemsize(x.dtype),
        min_mem=min_mem,
        max_mem=rechunker_max_mem,
    )

    for i, stage in enumerate(stages):
        last_stage = i == len(stages) - 1
        read_chunks, int_chunks, write_chunks = stage

        # Use target chunks for last stage
        target_chunks_ = target_chunks if last_stage else write_chunks

        if read_chunks == write_chunks:
            yield read_chunks, target_chunks_
        else:
            yield read_chunks, int_chunks
            yield write_chunks, target_chunks_


def _rechunk(x, copy_chunks, target_chunks):
    # rechunk x so that its target store has target_chunks, using copy_chunks as the size of chunks for copying from source to target

    normalized_copy_chunks = normalize_chunks(copy_chunks, x.shape, dtype=x.dtype)
    copy_chunks = to_chunksize(normalized_copy_chunks)

    copy_chunks_mem = array_memory(x.dtype, copy_chunks)

    target_chunks = normalize_chunks(target_chunks, x.shape, dtype=x.dtype)
    target_chunks = to_chunksize(target_chunks)

    def selection_function(out_key):
        out_coords = out_key[1:]
        return get_item(normalized_copy_chunks, out_coords)

    max_num_input_blocks = math.prod(
        math.ceil(c1 / c0) for c0, c1 in zip(x.chunksize, copy_chunks)
    )

    return map_selection(
        None,  # no function to apply after selection
        selection_function,
        x,
        x.shape,
        x.dtype,
        normalized_copy_chunks,
        max_num_input_blocks=max_num_input_blocks,
        target_chunks_=target_chunks,
        fusable_with_predecessors=False,
        fusable_with_successors=False,
        op_name="rechunk",
        extra_projected_mem=copy_chunks_mem,
    )


def merge_chunks(x, chunks):
    """Merge multiple chunks into one."""
    target_chunksize = chunks
    if len(target_chunksize) != x.ndim:
        raise ValueError(
            f"Chunks {target_chunksize} must have same number of dimensions as array ({x.ndim})"
        )
    if not all(c1 % c0 == 0 for c0, c1 in zip(x.chunksize, target_chunksize)):
        raise ValueError(
            f"Chunks {target_chunksize} must be a multiple of array's chunks {x.chunksize}"
        )

    target_chunks = normalize_chunks(chunks, x.shape, dtype=x.dtype)

    def selection_function(out_key):
        out_coords = out_key[1:]
        return get_item(target_chunks, out_coords)

    max_num_input_blocks = math.prod(
        c1 // c0 for c0, c1 in zip(x.chunksize, target_chunksize)
    )

    return map_selection(
        None,  # no function to apply after selection
        selection_function,
        x,
        x.shape,
        x.dtype,
        target_chunks,
        max_num_input_blocks=max_num_input_blocks,
    )


def reduction(
    x: "Array",
    func,
    combine_func=None,
    aggregate_func=None,
    axis=None,
    intermediate_dtype=None,
    dtype=None,
    keepdims=False,
    split_every=None,
    combine_sizes=None,
    extra_func_kwargs=None,
) -> "Array":
    """Apply a function to reduce an array along one or more axes.

    Parameters
    ----------
    x: Array
        Array being reduced along one or more axes.
    func: callable
        Function to apply to each chunk of data before reduction.
    combine_func: callable, optional
        Function which may be applied recursively to intermediate chunks of
        data. The number of chunks that are combined in each round is
        determined by the ``split_every`` parameter. The output of the
        function is a chunk with size one (or the size specified in
        ``combine_sizes``) in each of the reduction axes. If omitted,
        it defaults to ``func``.
    aggregate_func: callable, optional
        Function to apply to each of the final chunks to produce the final output.
    axis: int or sequence of ints, optional
        Axis or axes to aggregate upon. If omitted, aggregate along all axes.
    intermediate_dtype: dtype
        Data type of intermediate output.
    dtype: dtype
        Data type of output.
    keepdims: boolean, optional
        Whether the reduction function should preserve the reduced axes,
        or remove them.
    split_every: int >= 2 or dict(axis: int), optional
        The number of chunks to combine in one round along each axis in the
        recursive aggregation.
    combine_sizes: dict(axis: int), optional
        The resulting size of each axis after reduction. Each reduction axis
        defaults to size one if not specified.
    extra_func_kwargs: dict, optional
        Extra keyword arguments to pass to ``func`` and ``combine_func``.
    """
    if combine_func is None:
        if func is None:
            raise ValueError(
                "At least one of `func` and `combine_func` must be specified in reduction"
            )
        combine_func = func
    if axis is None:
        axis = tuple(range(x.ndim))
    if isinstance(axis, Integral):
        axis = (axis,)
    axis = validate_axis(axis, x.ndim)
    if intermediate_dtype is None:
        intermediate_dtype = dtype

    split_every = _normalize_split_every(split_every, axis)

    if func is None:
        initial_func = None
    else:
        initial_func = partial(
            func, axis=axis, keepdims=True, **(extra_func_kwargs or {})
        )
    result = partial_reduce(
        x,
        partial(combine_func, **(extra_func_kwargs or {})),
        initial_func=initial_func,
        split_every=split_every,
        dtype=intermediate_dtype,
        combine_sizes=combine_sizes,
    )

    # combine intermediates
    result = tree_reduce(
        result,
        partial(combine_func, **(extra_func_kwargs or {})),
        axis=axis,
        dtype=intermediate_dtype,
        split_every=split_every,
        combine_sizes=combine_sizes,
    )

    # aggregate final chunks
    if aggregate_func is not None:
        result = map_blocks(
            partial(aggregate_func, **(extra_func_kwargs or {})), result, dtype=dtype
        )

    if not keepdims:
        axis_to_squeeze = tuple(i for i in axis if result.shape[i] == 1)
        if len(axis_to_squeeze) > 0:
            result = squeeze(result, axis_to_squeeze)

    from cubed.array_api import astype

    result = astype(result, dtype, copy=False)

    return result


def _normalize_split_every(split_every, axis):
    split_every = split_every or 4
    if isinstance(split_every, dict):
        split_every = {k: split_every.get(k, 2) for k in axis}
    elif isinstance(split_every, Integral):
        n = builtins.max(int(split_every ** (1 / (len(axis) or 1))), 2)
        split_every = dict.fromkeys(axis, n)
    else:
        raise ValueError("split_every must be a int or a dict")
    return split_every


def tree_reduce(
    x,
    func,
    axis,
    dtype,
    split_every=None,
    combine_sizes=None,
):
    """Apply a reduction function repeatedly across multiple axes."""
    if axis is None:
        axis = tuple(range(x.ndim))
    if isinstance(axis, Integral):
        axis = (axis,)
    axis = validate_axis(axis, x.ndim)

    split_every = _normalize_split_every(split_every, axis)

    depth = 0
    for i, n in enumerate(x.numblocks):
        if i in split_every and split_every[i] != 1:
            depth = int(builtins.max(depth, math.ceil(math.log(n, split_every[i]))))
    for _ in range(depth):
        x = partial_reduce(
            x,
            func,
            split_every=split_every,
            dtype=dtype,
            combine_sizes=combine_sizes,
        )
    return x


def partial_reduce(
    x,
    func,
    initial_func=None,
    split_every=None,
    dtype=None,
    combine_sizes=None,
):
    """Apply a reduction function to multiple blocks across multiple axes.

    Parameters
    ----------
    x: Array
        Array being reduced along one or more axes
    func: callable
        Reduction function to apply to each chunk of data, resulting in a chunk
        with size one (or the size specified in ``combine_sizes``) in each of
        the reduction axes.
    initial_func: callable, optional
        Function to apply to each chunk of data before reduction.
    split_every: int >= 2 or dict(axis: int), optional
        The number of chunks to combine in one round along each axis in the
        recursive aggregation.
    dtype: dtype
        Output data type.
    combine_sizes: dict(axis: int), optional
        The resulting size of each axis after reduction. Each reduction axis
        defaults to size one if not specified.
    """
    # map over output chunks
    axis = tuple(ax for ax in split_every.keys())
    combine_sizes = combine_sizes or {}
    combine_sizes = {k: combine_sizes.get(k, 1) for k in axis}
    chunks = tuple(
        (combine_sizes[i],) * math.ceil(len(c) / split_every[i])
        if i in split_every
        else c
        for (i, c) in enumerate(x.chunks)
    )
    shape = tuple(map(sum, chunks))

    def key_function(out_key):
        out_coords = out_key[1:]

        # return a tuple with a single item that is an iterator of input keys to be merged
        in_keys = [
            list(
                range(
                    bi * split_every.get(i, 1),
                    min((bi + 1) * split_every.get(i, 1), x.numblocks[i]),
                )
            )
            for i, bi in enumerate(out_coords)
        ]
        return (iter([(x.name,) + tuple(p) for p in product(*in_keys)]),)

    # Since key_function returns an iterator of input keys, the the array chunks passed to
    # _partial_reduce are retrieved one at a time. However, we need an extra chunk of memory
    # to stay within limits (maybe because the iterator doesn't free the previous object
    # before getting the next). We also need extra memory to hold two reduced chunks, since
    # they are concatenated two at a time.
    extra_projected_mem = x.chunkmem + 2 * array_memory(dtype, to_chunksize(chunks))

    return general_blockwise(
        _partial_reduce,
        key_function,
        x,
        shapes=[shape],
        dtypes=[dtype],
        chunkss=[chunks],
        extra_projected_mem=extra_projected_mem,
        num_input_blocks=(sum(split_every.values()),),
        iterable_input_blocks=(True,),
        reduce_func=func,
        initial_func=initial_func,
        axis=axis,
    )


def _partial_reduce(arrays, reduce_func=None, initial_func=None, axis=None):
    # reduce each array in turn, accumulating in result
    assert not isinstance(arrays, list), (
        "partial reduce expects an iterator of array blocks, not a list"
    )
    result = None
    for array in arrays:
        if initial_func is not None:
            array = initial_func(array)
        reduced_chunk = reduce_func(array, axis=axis, keepdims=True)
        if result is None:
            result = reduced_chunk
        elif isinstance(result, dict):
            assert result.keys() == reduced_chunk.keys()
            result = {
                # only need to concatenate along first axis
                k: nxp.concat([result[k], reduced_chunk[k]], axis=axis[0])
                for k in result.keys()
            }
            result = reduce_func(result, axis=axis, keepdims=True)
        else:
            # only need to concatenate along first axis
            result = nxp.concat([result, reduced_chunk], axis=axis[0])
            result = reduce_func(result, axis=axis, keepdims=True)

    return result


def arg_reduction(x, /, arg_func, axis=None, *, keepdims=False, split_every=None):
    """A reduction that returns the array indexes, not the values."""
    dtype = nxp.__array_namespace_info__().default_dtypes(device=x.device)["indexing"]
    intermediate_dtype = [("i", dtype), ("v", x.dtype)]

    # initial map does arg reduction on each block, and uses block id to find the absolute index within whole array
    chunks = tuple((1,) * len(c) if i == axis else c for i, c in enumerate(x.chunks))
    out = map_blocks(
        partial(_arg_map_func, arg_func=arg_func),
        x,
        dtype=intermediate_dtype,
        chunks=chunks,
        axis=axis,
        size=to_chunksize(x.chunks)[axis],
    )

    # then reduce across blocks
    return reduction(
        out,
        _arg_func,
        combine_func=partial(_arg_combine, arg_func=arg_func),
        aggregate_func=_arg_aggregate,
        axis=axis,
        intermediate_dtype=intermediate_dtype,
        dtype=dtype,
        keepdims=keepdims,
        split_every=split_every,
    )


def _arg_map_func(a, axis, arg_func=None, size=None, block_id=None):
    i = arg_func(a, axis=axis, keepdims=True)
    # note that the array API doesn't have take_along_axis, so this may fail
    v = nxp.take_along_axis(a, i, axis=axis)
    # add block offset to i so it is absolute index within whole array
    offset = block_id[axis] * size
    return {"i": i + offset, "v": v}


def _arg_func(a, **kwargs):
    # pass through
    return {"i": a["i"], "v": a["v"]}


def _arg_combine(a, arg_func=None, **kwargs):
    # convert axis from single value tuple to int
    axis = kwargs.pop("axis")[0]

    i = a["i"]
    v = a["v"]

    # find indexes of values in v and apply to i and v
    vi = arg_func(v, axis=axis, **kwargs)
    # note that the array API doesn't have take_along_axis, so this may fail
    i_combined = nxp.take_along_axis(i, vi, axis=axis)
    v_combined = nxp.take_along_axis(v, vi, axis=axis)
    return {"i": i_combined, "v": v_combined}


def _arg_aggregate(a):
    # just return index values
    return a["i"]


def squeeze(x, /, axis):
    if not isinstance(axis, tuple):
        axis = (axis,)

    if any(x.shape[i] != 1 for i in axis):
        raise ValueError("cannot squeeze axis with size other than one")

    axis = validate_axis(axis, x.ndim)

    chunks = tuple(c for i, c in enumerate(x.chunks) if i not in axis)

    return map_blocks(
        nxp.squeeze, x, dtype=x.dtype, chunks=chunks, drop_axis=axis, axis=axis
    )


def unify_chunks(*args: "Array", **kwargs):
    # From dask unify_chunks
    if not args:
        return {}, []

    arginds = [(a, ind) for a, ind in partition(2, args)]  # [x, ij, y, jk]

    arrays, inds = zip(*arginds)
    if all(ind is None for ind in inds):
        return {}, list(arrays)
    if all(ind == inds[0] for ind in inds) and all(
        a.chunks == arrays[0].chunks for a in arrays
    ):
        return dict(zip(inds[0], arrays[0].chunks)), arrays

    nameinds = []
    blockdim_dict = dict()
    max_parts = 0
    for a, ind in arginds:
        if ind is not None:
            nameinds.append((a.name, ind))
            blockdim_dict[a.name] = a.chunks
            max_parts = max(max_parts, a.npartitions)
        else:
            nameinds.append((a, ind))

    chunkss = broadcast_dimensions(
        nameinds, blockdim_dict, consolidate=smallest_blockdim
    )

    arrays = []
    for a, i in arginds:
        if i is None:
            arrays.append(a)
        else:
            chunks = tuple(
                chunkss[j]
                if a.shape[n] > 1
                else (a.shape[n],)
                if not np.isnan(sum(chunkss[j]))
                else None
                for n, j in enumerate(i)
            )
            if chunks != a.chunks and all(a.chunks):
                # this will raise if chunks are not regular
                # but this should never happen with smallest_blockdim
                chunksize = to_chunksize(chunks)  # type: ignore
                arrays.append(rechunk(a, chunksize))
            else:
                arrays.append(a)
    return chunkss, arrays


def smallest_blockdim(blockdims):
    """Find the smallest block dimensions from the list of block dimensions

    Unlike Dask's common_blockdim, this returns regular chunks (assuming
    regular chunks are passed in).
    """
    if not any(blockdims):
        return ()
    non_trivial_dims = {d for d in blockdims if len(d) > 1}
    if len(non_trivial_dims) == 1:
        return first(non_trivial_dims)
    if len(non_trivial_dims) == 0:
        return max(blockdims, key=first)

    if len(set(map(sum, non_trivial_dims))) > 1:
        raise ValueError("Chunks do not add up to same value", blockdims)

    # find dims with the smallest first chunk
    m = -1
    out = None
    for ntd in non_trivial_dims:
        if m == -1 or ntd[0] < m:
            m = ntd[0]
            out = ntd
    return out


def scan(
    array: "Array",
    func: Callable,
    *,
    preop: Callable,
    binop: Callable,
    axis: int,
    dtype=None,
    include_initial=False,
    split_every: int = 5,
) -> "Array":
    """
    Generic parallel scan.

    Parameters
    ----------
    x: Cubed Array
    func: callable
        Scan or cumulative function like np.cumsum or np.cumprod
    preop: callable
        Function applied blockwise that reduces each block to a single value
        along ``axis``. For ``np.cumsum`` this is ``np.sum`` and for ``np.cumprod`` this is ``np.prod``.
    binop: callable
        Associated binary operator like ``np.cumsum->add`` or ``np.cumprod->mul``
    axis: int
    dtype: dtype
    include_initial: bool
        Whether to include the identity value as the first value in the output.

    Notes
    -----
    This method uses a variant of the Blelloch (1989) algorithm.

    Returns
    -------
    Array

    See also
    --------
    cumsum
    cumprod
    """

    # Note that if include_initial=True the final value is *not* included in the output.
    # To include the final value is tricky with constant chunk sizes, since if the last
    # chunk is full then a new chunk of size one needs to be added for the final value.
    # TODO: add an include_final argument (default True)

    axis = validate_axis(axis, array.ndim)

    # Blelloch (1990) out-of-core algorithm.
    # 1. First, scan blockwise
    scanned = map_blocks(
        func, array, dtype=dtype, axis=axis, include_initial=include_initial
    )
    # If there is only a single chunk, we can be done
    if array.numblocks[axis] == 1:
        return scanned

    # 2. Calculate the reduction using `preop`
    #    Use `partial_reduce` to also merge to a decent intermediate chunksize
    #    since reduced.chunksize[axis] == 1

    def identity_func(a, **kwargs):
        return a

    split_size = min(split_every, array.numblocks[axis])
    reduced = partial_reduce(
        array,
        initial_func=partial(preop, axis=axis, keepdims=True),
        func=identity_func,
        split_every={axis: split_size},
        dtype=dtype,
        combine_sizes={axis: split_size},
    )

    # 3. Now scan `reduced` to generate the increments for each block of `scanned`.
    #    Here we diverge from Blelloch, who runs a balanced tree algorithm to calculate the scan.
    #    Instead we generalize recursively apply the scan to `reduced`.
    #    Note we always want to include the initial identity value (but not the final value)
    #    so blocks line up correctly.
    increment = scan(
        reduced,
        func,
        preop=preop,
        binop=binop,
        axis=axis,
        dtype=dtype,
        include_initial=True,
    )

    # 4. Back to Blelloch. Now that we have the increment, add it to the blocks of `scanned`.
    #    Use general_blockwise with a key function since the chunks of increment and scanned aren't aligned anymore.
    assert increment.shape[axis] == scanned.numblocks[axis]

    def key_function(out_key):
        out_coords = out_key[1:]
        inc_coords = tuple(
            bi // split_every if i == axis else bi for i, bi in enumerate(out_coords)
        )
        return ((scanned.name,) + out_coords, (increment.name,) + inc_coords)

    def _scan_binop(scn, inc, block_id=None, **kwargs):
        bi = block_id[axis] % split_every
        ind = tuple(
            slice(bi, bi + 1) if i == axis else slice(None) for i in range(inc.ndim)
        )
        return binop(scn, inc[ind])

    # 5. Bada-bing, bada-boom.
    out = general_blockwise(
        _scan_binop,
        key_function,
        scanned,
        increment,
        shapes=[scanned.shape],
        dtypes=[scanned.dtype],
        chunkss=[scanned.chunks],
        extra_projected_mem=scanned.chunkmem * 2,  # arbitrary
    )

    from cubed import Array

    assert isinstance(out, Array)  # single output
    return out



import math
import operator

import numpy as np

from cubed.array_api.creation_functions import asarray
from cubed.array_api.data_type_functions import result_type
from cubed.array_api.dtypes import (
    _boolean_dtypes,
    _complex_floating_dtypes,
    _dtype_categories,
    _floating_dtypes,
    _integer_dtypes,
    _integer_or_boolean_dtypes,
    _numeric_dtypes,
    complex64,
    complex128,
    float32,
    float64,
)
from cubed.array_api.linear_algebra_functions import matmul
from cubed.backend_array_api import namespace as nxp
from cubed.core.array import CoreArray
from cubed.core.ops import elemwise
from cubed.utils import itemsize, memory_repr
from cubed.vendor.dask.widgets import get_template

ARRAY_SVG_SIZE = (
    120  # cubed doesn't have a config module like dask does so hard-code this for now
)


class Array(CoreArray):
    """Chunked array backed by Zarr storage that conforms to the Python Array API standard."""

    def __init__(self, name, zarray, spec, plan):
        super().__init__(name, zarray, spec, plan)

    def __array__(self, dtype=None) -> np.ndarray:
        x = self.compute()
        if dtype and x.dtype != dtype:
            x = x.astype(dtype)
        if not isinstance(x, np.ndarray):
            x = np.array(x)
        return x

    def __repr__(self):
        return f"cubed.Array<{self.name}, shape={self.shape}, dtype={self.dtype}, chunks={self.chunks}>"

    def _repr_html_(self):
        try:
            grid = self.to_svg(size=ARRAY_SVG_SIZE)
        except NotImplementedError:
            grid = ""

        if not math.isnan(self.nbytes):
            nbytes = memory_repr(self.nbytes)
            cbytes = memory_repr(math.prod(self.chunksize) * itemsize(self.dtype))
        else:
            nbytes = "unknown"
            cbytes = "unknown"

        return get_template("array.html.j2").render(
            array=self,
            grid=grid,
            nbytes=nbytes,
            cbytes=cbytes,
            arrs_in_plan=f"{self.plan._finalize().num_arrays()} arrays in Plan",
            arrtype="np.ndarray",
        )

    def to_svg(self, size=500):
        """Convert chunks from Cubed Array into an SVG Image

        Parameters
        ----------
        chunks: tuple
        size: int
            Rough size of the image

        Examples
        --------
        >>> x.to_svg(size=500)  # doctest: +SKIP

        Returns
        -------
        text: An svg string depicting the array as a grid of chunks
        """
        from cubed.vendor.dask.array.svg import svg

        return svg(self.chunks, size=size)

    def _repr_inline_(self, max_width):
        """
        Format to a single line with at most max_width characters. Used by xarray.
        """
        return f"cubed.Array<chunksize={self.chunksize}>"

    # Attributes

    @property
    def device(self):
        return "cpu"

    @property
    def mT(self):
        from cubed.array_api.linear_algebra_functions import matrix_transpose

        return matrix_transpose(self)

    @property
    def T(self):
        if self.ndim != 2:
            raise ValueError("x.T requires x to have 2 dimensions.")
        from cubed.array_api.linear_algebra_functions import matrix_transpose

        return matrix_transpose(self)

    # Arithmetic Operators

    def __neg__(self, /):
        if self.dtype not in _numeric_dtypes:
            raise TypeError("Only numeric dtypes are allowed in __neg__")
        return elemwise(nxp.negative, self, dtype=self.dtype)

    def __pos__(self, /):
        if self.dtype not in _numeric_dtypes:
            raise TypeError("Only numeric dtypes are allowed in __pos__")
        return elemwise(nxp.positive, self, dtype=self.dtype)

    def __add__(self, other, /):
        other = self._check_allowed_dtypes(other, "numeric", "__add__")
        if other is NotImplemented:
            return other
        return elemwise(nxp.add, self, other, dtype=result_type(self, other))

    def __sub__(self, other, /):
        other = self._check_allowed_dtypes(other, "numeric", "__sub__")
        if other is NotImplemented:
            return other
        return elemwise(nxp.subtract, self, other, dtype=result_type(self, other))

    def __mul__(self, other, /):
        other = self._check_allowed_dtypes(other, "numeric", "__mul__")
        if other is NotImplemented:
            return other
        return elemwise(nxp.multiply, self, other, dtype=result_type(self, other))

    def __truediv__(self, other, /):
        other = self._check_allowed_dtypes(other, "floating-point", "__truediv__")
        if other is NotImplemented:
            return other
        return elemwise(nxp.divide, self, other, dtype=result_type(self, other))

    def __floordiv__(self, other, /):
        other = self._check_allowed_dtypes(other, "real numeric", "__floordiv__")
        if other is NotImplemented:
            return other
        return elemwise(nxp.floor_divide, self, other, dtype=result_type(self, other))

    def __mod__(self, other, /):
        other = self._check_allowed_dtypes(other, "real numeric", "__mod__")
        if other is NotImplemented:
            return other
        return elemwise(nxp.remainder, self, other, dtype=result_type(self, other))

    def __pow__(self, other, /):
        other = self._check_allowed_dtypes(other, "numeric", "__pow__")
        if other is NotImplemented:
            return other
        return elemwise(nxp.pow, self, other, dtype=result_type(self, other))

    # Array Operators

    def __matmul__(self, other, /):
        other = self._check_allowed_dtypes(other, "numeric", "__matmul__")
        if other is NotImplemented:
            return other
        return matmul(self, other)

    # Bitwise Operators

    def __invert__(self, /):
        if self.dtype not in _integer_or_boolean_dtypes:
            raise TypeError("Only integer or boolean dtypes are allowed in __invert__")
        return elemwise(nxp.bitwise_invert, self, dtype=self.dtype)

    def __and__(self, other, /):
        other = self._check_allowed_dtypes(other, "integer or boolean", "__and__")
        if other is NotImplemented:
            return other
        return elemwise(nxp.bitwise_and, self, other, dtype=result_type(self, other))

    def __or__(self, other, /):
        other = self._check_allowed_dtypes(other, "integer or boolean", "__or__")
        if other is NotImplemented:
            return other
        return elemwise(nxp.bitwise_or, self, other, dtype=result_type(self, other))

    def __xor__(self, other, /):
        other = self._check_allowed_dtypes(other, "integer or boolean", "__xor__")
        if other is NotImplemented:
            return other
        return elemwise(nxp.bitwise_xor, self, other, dtype=result_type(self, other))

    def __lshift__(self, other, /):
        other = self._check_allowed_dtypes(other, "integer", "__lshift__")
        if other is NotImplemented:
            return other
        return elemwise(
            nxp.bitwise_left_shift, self, other, dtype=result_type(self, other)
        )

    def __rshift__(self, other, /):
        other = self._check_allowed_dtypes(other, "integer", "__rshift__")
        if other is NotImplemented:
            return other
        return elemwise(
            nxp.bitwise_right_shift, self, other, dtype=result_type(self, other)
        )

    # Comparison Operators

    def __eq__(self, other, /):
        other = self._check_allowed_dtypes(other, "all", "__eq__")
        if other is NotImplemented:
            return other
        return elemwise(nxp.equal, self, other, dtype=nxp.bool)

    def __ge__(self, other, /):
        other = self._check_allowed_dtypes(other, "all", "__ge__")
        if other is NotImplemented:
            return other
        return elemwise(nxp.greater_equal, self, other, dtype=nxp.bool)

    def __gt__(self, other, /):
        other = self._check_allowed_dtypes(other, "all", "__gt__")
        if other is NotImplemented:
            return other
        return elemwise(nxp.greater, self, other, dtype=nxp.bool)

    def __le__(self, other, /):
        other = self._check_allowed_dtypes(other, "all", "__le__")
        if other is NotImplemented:
            return other
        return elemwise(nxp.less_equal, self, other, dtype=nxp.bool)

    def __lt__(self, other, /):
        other = self._check_allowed_dtypes(other, "all", "__lt__")
        if other is NotImplemented:
            return other
        return elemwise(nxp.less, self, other, dtype=nxp.bool)

    def __ne__(self, other, /):
        other = self._check_allowed_dtypes(other, "all", "__ne__")
        if other is NotImplemented:
            return other
        return elemwise(nxp.not_equal, self, other, dtype=nxp.bool)

    # Reflected Operators

    # (Reflected) Arithmetic Operators

    def __radd__(self, other, /):
        other = self._check_allowed_dtypes(other, "numeric", "__radd__")
        if other is NotImplemented:
            return other
        return elemwise(nxp.add, other, self, dtype=result_type(self, other))

    def __rsub__(self, other, /):
        other = self._check_allowed_dtypes(other, "numeric", "__rsub__")
        if other is NotImplemented:
            return other
        return elemwise(nxp.subtract, other, self, dtype=result_type(self, other))

    def __rmul__(self, other, /):
        other = self._check_allowed_dtypes(other, "numeric", "__rmul__")
        if other is NotImplemented:
            return other
        return elemwise(nxp.multiply, other, self, dtype=result_type(self, other))

    def __rtruediv__(self, other, /):
        other = self._check_allowed_dtypes(other, "floating-point", "__rtruediv__")
        if other is NotImplemented:
            return other
        return elemwise(nxp.divide, other, self, dtype=result_type(self, other))

    def __rfloordiv__(self, other, /):
        other = self._check_allowed_dtypes(other, "numeric", "__rfloordiv__")
        if other is NotImplemented:
            return other
        return elemwise(nxp.floor_divide, other, self, dtype=result_type(self, other))

    def __rmod__(self, other, /):
        other = self._check_allowed_dtypes(other, "numeric", "__rmod__")
        if other is NotImplemented:
            return other
        return elemwise(nxp.remainder, other, self, dtype=result_type(self, other))

    def __rpow__(self, other, /):
        other = self._check_allowed_dtypes(other, "numeric", "__rpow__")
        if other is NotImplemented:
            return other
        return elemwise(nxp.pow, other, self, dtype=result_type(self, other))

    # (Reflected) Array Operators

    def __rmatmul__(self, other, /):
        other = self._check_allowed_dtypes(other, "numeric", "__rmatmul__")
        if other is NotImplemented:
            return other
        return matmul(other, self)

    # (Reflected) Bitwise Operators

    def __rand__(self, other, /):
        other = self._check_allowed_dtypes(other, "integer or boolean", "__rand__")
        if other is NotImplemented:
            return other
        return elemwise(nxp.bitwise_and, other, self, dtype=result_type(self, other))

    def __ror__(self, other, /):
        other = self._check_allowed_dtypes(other, "integer or boolean", "__ror__")
        if other is NotImplemented:
            return other
        return elemwise(nxp.bitwise_or, other, self, dtype=result_type(self, other))

    def __rxor__(self, other, /):
        other = self._check_allowed_dtypes(other, "integer or boolean", "__rxor__")
        if other is NotImplemented:
            return other
        return elemwise(nxp.bitwise_xor, other, self, dtype=result_type(self, other))

    def __rlshift__(self, other, /):
        other = self._check_allowed_dtypes(other, "integer", "__rlshift__")
        if other is NotImplemented:
            return other
        return elemwise(
            nxp.bitwise_left_shift, other, self, dtype=result_type(self, other)
        )

    def __rrshift__(self, other, /):
        other = self._check_allowed_dtypes(other, "integer", "__rrshift__")
        if other is NotImplemented:
            return other
        return elemwise(
            nxp.bitwise_right_shift, other, self, dtype=result_type(self, other)
        )

    # Methods

    def __abs__(self, /):
        if self.dtype not in _numeric_dtypes:
            raise TypeError("Only numeric dtypes are allowed in __abs__")
        if self.dtype == complex64:
            dtype = float32
        elif self.dtype == complex128:
            dtype = float64
        else:
            dtype = self.dtype
        return elemwise(nxp.abs, self, dtype=dtype)

    def __array_namespace__(self, /, *, api_version=None):
        if api_version is not None and api_version not in (
            "2021.12",
            "2022.12",
            "2023.12",
            "2024.12",
        ):
            raise ValueError(f"Unrecognized array API version: {api_version!r}")
        import cubed

        return cubed

    def __bool__(self, /):
        if self.ndim != 0:
            raise TypeError("bool is only allowed on arrays with 0 dimensions")
        return bool(self.compute())

    def __complex__(self, /):
        if self.ndim != 0:
            raise TypeError("complex is only allowed on arrays with 0 dimensions")
        return complex(self.compute())

    def __float__(self, /):
        if self.ndim != 0:
            raise TypeError("float is only allowed on arrays with 0 dimensions")
        if self.dtype in _complex_floating_dtypes:
            raise TypeError("float is not allowed on complex floating-point arrays")
        return float(self.compute())

    def __index__(self, /):
        if self.ndim != 0:
            raise TypeError("index is only allowed on arrays with 0 dimensions")
        return operator.index(self.compute())

    def __int__(self, /):
        if self.ndim != 0:
            raise TypeError("int is only allowed on arrays with 0 dimensions")
        if self.dtype in _complex_floating_dtypes:
            raise TypeError("int is not allowed on complex floating-point arrays")
        return int(self.compute())

    # Utility methods

    def _check_allowed_dtypes(self, other, dtype_category, op):
        if (
            dtype_category != "all"
            and self.dtype not in _dtype_categories[dtype_category]
        ):
            raise TypeError(f"Only {dtype_category} dtypes are allowed in {op}")
        if isinstance(other, (int, complex, float, bool)):
            other = self._promote_scalar(other)
        elif isinstance(other, CoreArray):
            if (
                dtype_category != "all"
                and other.dtype not in _dtype_categories[dtype_category]
            ):
                raise TypeError(f"Only {dtype_category} dtypes are allowed in {op}")
        else:
            return NotImplemented

        # TODO: more from numpy.array_api

        return other

    def _promote_scalar(self, scalar):
        if isinstance(scalar, bool):
            if self.dtype not in _boolean_dtypes:
                raise TypeError(
                    "Python bool scalars can only be promoted with bool arrays"
                )
        elif isinstance(scalar, int):
            if self.dtype in _boolean_dtypes:
                raise TypeError(
                    "Python int scalars cannot be promoted with bool arrays"
                )
            if self.dtype in _integer_dtypes:
                info = nxp.iinfo(self.dtype)
                if not (info.min <= scalar <= info.max):
                    raise OverflowError(
                        "Python int scalars must be within the bounds of the dtype for integer arrays"
                    )
            # int + array(floating) is allowed
        elif isinstance(scalar, float):
            if self.dtype not in _floating_dtypes:
                raise TypeError(
                    "Python float scalars can only be promoted with floating-point arrays."
                )
        elif isinstance(scalar, complex):
            if self.dtype not in _complex_floating_dtypes:
                raise TypeError(
                    "Python complex scalars can only be promoted with complex floating-point arrays."
                )
        else:
            raise TypeError("'scalar' must be a Python scalar")
        return asarray(scalar, dtype=self.dtype, spec=self.spec)
